# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c) 2015 Jean-Philippe Braun <eon@patapon.info>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from hashlib import md5

from lollypop.define import App, StorageType, ScanUpdate, Type
from lollypop.objects_track import Track
from lollypop.objects import Base
from lollypop.utils import emit_signal
from lollypop.collection_item import CollectionItem
from lollypop.logger import Logger


class Disc:
    """
        Represent an album disc
    """

    def __init__(self, album, disc_number, storage_type, skipped):
        self.db = App().albums
        self.__tracks = []
        self.__album = album
        self.__storage_type = storage_type
        self.__number = disc_number
        self.__skipped = skipped

    def __del__(self):
        """
            Remove ref cycles
        """
        self.__album = None

    # Used by pickle
    def __getstate__(self):
        self.db = None
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.db = App().albums

    def set_tracks(self, tracks):
        """
            Set disc tracks
            @param tracks as [Track]
        """
        self.__tracks = tracks

    @property
    def number(self):
        """
            Get disc number
        """
        return self.__number

    @property
    def album(self):
        """
            Get disc album
            @return Album
        """
        return self.__album

    @property
    def track_ids(self):
        """
            Get disc track ids
            @return [int]
        """
        return [track.id for track in self.tracks]

    @property
    def track_uris(self):
        """
            Get disc track uris
            @return [str]
        """
        return [track.uri for track in self.tracks]

    @property
    def tracks(self):
        """
            Get disc tracks
            @return [Track]
        """
        if not self.__tracks and self.album.id is not None:
            self.__tracks = [Track(track_id, self.album)
                             for track_id in self.db.get_disc_track_ids(
                                    self.album.id,
                                    self.album.genre_ids,
                                    self.album.artist_ids,
                                    self.number,
                                    self.__storage_type,
                                    self.__skipped)]
        return self.__tracks


class Album(Base):
    """
        Represent an album
    """
    DEFAULTS = {"artists": [],
                "artist_ids": [],
                "year": None,
                "timestamp": 0,
                "uri": "",
                "popularity": 0,
                "rate": 0,
                "mtime": 1,
                "synced": 0,
                "loved": False,
                "storage_type": 0,
                "mb_album_id": None,
                "lp_album_id": None}

    def __init__(self, album_id=None, genre_ids=[], artist_ids=[],
                 skipped=True):
        """
            Init album
            @param album_id as int
            @param genre_ids as [int]
            @param artist_ids as [int]
            @param skipped as bool
        """
        Base.__init__(self, App().albums)
        self.id = album_id
        self.genre_ids = genre_ids
        self.__tracks = []
        self.__discs = []
        self.__name = None
        self.__skipped = skipped
        self.__disc_number = None
        self.__original_year = Type.NONE
        self.__tracks_storage_type = self.storage_type
        # Use artist ids from db else
        if artist_ids:
            artists = []
            for artist_id in set(artist_ids) | set(self.artist_ids):
                artists.append(App().artists.get_name(artist_id))
            self.artists = artists
            self.artist_ids = artist_ids

    def __del__(self):
        """
            Remove ref cycles
        """
        self.reset_tracks()

    # Used by pickle
    def __getstate__(self):
        self.db = None
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.db = App().albums

    def set_discs(self, discs):
        """
            Set album discs
            @param discs as [Disc]
        """
        self.__discs = discs

    def set_disc_number(self, disc_number):
        """
            Set album disc
            @param disc_number as int
        """
        self.__original_year = Type.NONE
        self.__disc_number = disc_number

    def set_tracks(self, tracks, clone=True):
        """
            Set album tracks, do not disable clone if you know self is already
            used
            @param tracks as [Track]
            @param clone as bool
        """
        if clone:
            self.__tracks = []
            for track in tracks:
                new_track = Track(track.id, self)
                self.__tracks.append(new_track)
        # Album tracks already belong to self
        # Detach those tracks
        elif self.__tracks:
            new_album = Album(self.id, self.genre_ids, self.artist_ids)
            new__tracks = []
            for track in self.__tracks:
                if track not in tracks:
                    track.set_album(new_album)
                    new__tracks.append(track)
            new_album.__tracks = new__tracks
            self.__tracks = tracks
        else:
            self.__tracks = tracks

    def append_track(self, track, clone=True):
        """
            Append track to album.
            Clone: always do this if track is used in UI/Player
            @param track as Track
            @param clone as bool
        """
        if clone:
            self.__tracks.append(Track(track.id, self))
        else:
            self.__tracks.append(track)
            track.set_album(self)

    def append_tracks(self, tracks, clone=True):
        """
            Append tracks to album
            Clone: always do this if track is used in UI/Player
            @param tracks as [Track]
            @param clone as bool
        """
        for track in tracks:
            self.append_track(track, clone)

    def remove_track(self, track):
        """
            Remove track from album, album id is None if empty
            @param track as Track
        """
        for _track in self.tracks:
            if track.id == _track.id:
                self.__tracks.remove(_track)
        empty = len(self.__tracks) == 0
        if empty:
            # We don't the album to load tracks anymore
            self.id = None

    def reset_tracks(self):
        """
            Reset album tracks, useful for tracks loaded async
        """
        self.__tracks = []
        self.__discs = []
        self.reset("artists")
        self.reset("artist_ids")
        self.reset("lp_album_id")

    def disc_names(self, disc_number):
        """
            Disc names
            @param disc_number as int
            @return disc names as [str]
        """
        return self.db.get_disc_names(self.id, disc_number)

    def set_loved(self, loved):
        """
            Mark album as loved
            @param loved as bool
        """
        if self.id >= 0:
            self.db.set_loved(self.id, loved)
            self.loved = loved

    def set_uri(self, uri):
        """
            Set album uri
            @param uri as str
        """
        if self.id >= 0:
            self.db.set_uri(self.id, uri)
        self.uri = uri

    def get_track(self, track_id):
        """
            Get track
            @param track_id as int
            @return Track
        """
        for track in self.tracks:
            if track.id == track_id:
                return track
        return Track()

    def save(self, save):
        """
            Save album to collection.
            @param save as bool
        """
        # Save tracks
        for track_id in self.track_ids:
            if save:
                App().tracks.set_storage_type(track_id, StorageType.SAVED)
            else:
                App().tracks.set_storage_type(track_id, StorageType.EPHEMERAL)
        # Save album
        self.__save(save)

    def save_track(self, save, track):
        """
            Save track to collection
            @param save as bool
            @param track as Track
        """
        if save:
            App().tracks.set_storage_type(track.id, StorageType.SAVED)
        else:
            App().tracks.set_storage_type(track.id, StorageType.EPHEMERAL)
        # Save album
        self.__save(save)

    def load_tracks(self, cancellable):
        """
            Load album tracks from Spotify,
            do not call this for Storage.COLLECTION
            @param cancellable as Gio.Cancellable
            @return status as bool
        """
        try:
            if self.storage_type & (StorageType.COLLECTION |
                                    StorageType.EXTERNAL):
                return False
            elif self.synced != 0 and self.synced != len(self.tracks):
                from lollypop.search import Search
                Search().load_tracks(self, cancellable)
                self.reset_tracks()
        except Exception as e:
            Logger.warning("Album::load_tracks(): %s" % e)
        return True

    def set_synced(self, mask):
        """
            Set synced mask
            @param mask as int
        """
        self.db.set_synced(self.id, mask)
        self.synced = mask

    def clone(self, skipped):
        """
            Clone album
            @param skipped as bool
            @return album
        """
        album = Album(self.id, self.genre_ids, self.artist_ids, skipped)
        if skipped:
            album.set_tracks(self.tracks)
        return album

    def set_storage_type(self, storage_type):
        """
            Set storage type
            @param storage_type as StorageType
        """
        self.__tracks_storage_type = storage_type

    def set_skipped(self):
        """
            Set album as skipped, not allowing skipped tracks
        """
        self.__skipped = True

    def merge_discs(self):
        """
            Merge album discs
            @return Disc
        """
        self.__original_year = None
        tracks = self.tracks
        disc = Disc(self, 0, self.__tracks_storage_type, self.__skipped)
        disc.set_tracks(tracks)
        self.__discs = [disc]

    @property
    def original_year(self):
        """
            Get disc original year
            @return int/None
        """
        if self.__original_year == Type.NONE:
            self.__original_year = App().tracks.get_year_for_album(
                self.id, self.__disc_number)
        return self.__original_year

    @property
    def collection_item(self):
        """
            Get collection item related to album
            @return CollectionItem
        """
        item = CollectionItem(album_id=self.id,
                              album_name=self.name,
                              artist_ids=self.artist_ids,
                              lp_album_id=self.lp_album_id)
        return item

    @property
    def name(self):
        """
            Get album name
            @return str
        """
        if self.__name is not None:
            return self.__name
        if self.__disc_number is None:
            self.__name = self.db.get_name(self.id)
        else:
            disc_names = self.disc_names(self.__disc_number)
            if disc_names:
                self.__name = ", ".join(disc_names)
            else:
                self.__name = self.db.get_name(self.id)
        return self.__name

    @property
    def is_web(self):
        """
            True if track is a web track
            @return bool
        """
        return not self.storage_type & (StorageType.COLLECTION |
                                        StorageType.EXTERNAL)

    @property
    def tracks_count(self):
        """
            Get tracks count
            @return int
        """
        if self.__tracks:
            return len(self.__tracks)
        else:
            return self.db.get__tracks_count(
                self.id,
                self.genre_ids,
                self.artist_ids)

    @property
    def track_ids(self):
        """
            Get album track ids
            @return [int]
        """
        return [track.id for track in self.tracks]

    @property
    def track_uris(self):
        """
            Get album track uris
            @return [str]
        """
        return [track.uri for track in self.tracks]

    @property
    def tracks(self):
        """
            Get album tracks
            @return [Track]
        """
        if self.id is None:
            return []
        if self.__tracks:
            return self.__tracks
        tracks = []
        for disc in self.discs:
            tracks += disc.tracks
        # Already cached by another thread
        if not self.__tracks:
            self.__tracks = tracks
        return tracks

    @property
    def discs(self):
        """
            Get albums discs
            @return [Disc]
        """
        if self.__discs:
            return self.__discs
        discs = []
        if self.__disc_number is None:
            disc_numbers = self.db.get_discs(self.id)
        else:
            disc_numbers = [self.__disc_number]
        for disc_number in disc_numbers:
            disc = Disc(self, disc_number,
                        self.__tracks_storage_type,
                        self.__skipped)
            if disc.tracks:
                discs.append(disc)
        # Already cached by another thread
        if not self.__discs:
            self.__discs = discs
        return self.__discs

    @property
    def duration(self):
        """
            Get album duration and handle caching
            @return int
        """
        if self.__tracks:
            track_ids = [track.lp_track_id for track in self.tracks]
            track_str = "%s" % sorted(track_ids)
            track_hash = md5(track_str.encode("utf-8")).hexdigest()
            album_hash = "%s-%s-%s" % (
                self.lp_album_id, track_hash, self.__disc_number)
        else:
            album_hash = "%s-%s-%s-%s" % (self.lp_album_id,
                                          self.genre_ids,
                                          self.artist_ids,
                                          self.__disc_number)
        duration = App().cache.get_duration(album_hash)
        if duration is None:
            if self.__tracks:
                duration = 0
                for track in self.__tracks:
                    duration += track.duration
            else:
                duration = self.db.get_duration(self.id,
                                                self.genre_ids,
                                                self.artist_ids,
                                                self.__disc_number)
            App().cache.set_duration(self.id, album_hash, duration)
        return duration

#######################
# PRIVATE             #
#######################
    def __save(self, save):
        """
            Save album to collection.
            @param save as bool
        """
        # Save album by updating storage type
        if save:
            self.db.set_storage_type(self.id, StorageType.SAVED)
        else:
            self.db.set_storage_type(self.id, StorageType.EPHEMERAL)
        self.reset("mtime")
        if save:
            item = CollectionItem(artist_ids=self.artist_ids,
                                  album_id=self.id)
            emit_signal(App().scanner, "updated", item,
                        ScanUpdate.ADDED)
        else:
            removed_artist_ids = []
            for artist_id in self.artist_ids:
                if not App().artists.get_name(artist_id):
                    removed_artist_ids.append(artist_id)
            item = CollectionItem(artist_ids=removed_artist_ids,
                                  album_id=self.id)
            emit_signal(App().scanner, "updated", item,
                        ScanUpdate.REMOVED)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib


class IndicatorWidget(Gtk.EventBox):
    """
        Show play/loved indicator
        If embedded in a Popover, will not affect playlists but only player
        playlists
    """

    def __init__(self, row, view_type):
        """
            Init indicator widget, ui will be set when needed
            @param row as Row
            @param view_type as ViewType
        """
        Gtk.EventBox.__init__(self)
        self.__row = row
        self.__view_type = view_type
        self.__pass = 1
        self.__timeout_id = None
        self.__stack = None
        self.connect("destroy", self.__on_destroy)
        # min-width = 24px, borders = 2px, padding = 8px
        self.set_size_request(34, -1)
        self.__stack = Gtk.Stack()
        self.__stack.set_transition_duration(500)
        self.__stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.__stack.show()
        self.add(self.__stack)

    def play(self):
        """
            Show play indicator
        """
        self.__init_play()
        self.__stack.set_visible_child_name("play")

    def loved(self):
        """
            Show loved indicator
        """
        self.__init_loved()
        self.__loved.set_from_icon_name("emblem-favorite-symbolic",
                                        Gtk.IconSize.MENU)
        self.__stack.set_visible_child_name("loved")

    def skip(self):
        """
            Show skip indicator
        """
        self.__init_loved()
        self.__loved.set_from_icon_name("media-skip-forward-symbolic",
                                        Gtk.IconSize.MENU)
        self.__stack.set_visible_child_name("loved")

    def play_loved(self):
        """
            Show play/loved indicator
        """
        self.__init_play()
        self.__init_loved()
        self.__pass = 1
        self.play()
        self.__timeout_id = GLib.timeout_add(1000, self.__play_loved)

    def load(self):
        """
            Show load indicator
        """
        self.__init_load()
        self.__stack.set_visible_child_name("load")
        self.__stack.get_visible_child().start()

    def clear(self):
        """
            Clear widget
        """
        load = self.__stack.get_child_by_name("load")
        if load is not None:
            load.stop()
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None

#######################
# PRIVATE             #
#######################
    def __init_play(self):
        """
            Init play button
        """
        if self.__stack.get_child_by_name("play") is not None:
            return
        play = Gtk.Image.new_from_icon_name("media-playback-start-symbolic",
                                            Gtk.IconSize.MENU)
        play.show()
        self.__stack.add_named(play, "play")

    def __init_loved(self):
        """
            Init loved button
        """
        if self.__stack.get_child_by_name("loved") is not None:
            return
        self.__loved = Gtk.Image()
        self.__loved.show()
        self.__stack.add_named(self.__loved, "loved")

    def __init_load(self):
        """
            Init load spinner
        """
        if self.__stack.get_child_by_name("load") is not None:
            return
        spinner = Gtk.Spinner()
        spinner.show()
        self.__stack.add_named(spinner, "load")

    def __on_destroy(self, widget):
        """
            Clear timeout
            @param widget as Gtk.Widget
        """
        self.__row = None
        self.clear()

    def __play_loved(self):
        """
            Show play/loved indicator
        """
        if self.__timeout_id is None:
            return False
        if self.__stack.get_visible_child_name() == "play":
            if self.__pass == 5:
                self.__pass = 0
                self.loved()
        else:
            self.play()
        self.__pass += 1
        return True



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from lollypop.define import App, ArtSize, ViewType
from lollypop.widgets_banner import BannerWidget
from lollypop.utils import popup_widget


class SearchBannerWidget(BannerWidget):
    """
        Banner for search
    """

    def __init__(self):
        """
            Init banner
        """
        BannerWidget.__init__(self, ViewType.OVERLAY)
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/SearchBannerWidget.ui")
        self.__menu = builder.get_object("menu")
        self.__spinner = builder.get_object("spinner")
        self.__entry = builder.get_object("entry")
        widget = builder.get_object("widget")
        self._overlay.add_overlay(widget)
        self._overlay.set_overlay_pass_through(widget, True)
        self.connect("map", self.__on_map)
        builder.connect_signals(self)

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        BannerWidget.update_for_width(self, width)
        self.__entry.set_size_request(self.width / 3, -1)

    @property
    def spinner(self):
        """
            Get banner spinner
            @return Gtk.Spinner
        """
        return self.__spinner

    @property
    def entry(self):
        """
            Get banner entry
            @return Gtk.Entry
        """
        return self.__entry

    @property
    def height(self):
        """
            Get wanted height
            @return int
        """
        return ArtSize.SMALL

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update entry width
            @param allocation as Gtk.Allocation
        """
        if BannerWidget._handle_width_allocate(self, allocation):
            self.__entry.set_size_request(self.width / 3, -1)

    def _on_menu_button_clicked(self, button):
        """
            Show playlist menu
            @param button as Gtk.Button
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_search import SearchMenu
        menu = SearchMenu(App().window.folded)
        menu_widget = MenuBuilder(menu)
        menu_widget.show()
        popup_widget(menu_widget, button, None, None, button)

#######################
# PRIVATE             #
#######################
    def __on_map(self, widget):
        """
            Grab focus
            @param widget as Gtk.Widget
        """
        GLib.idle_add(self.__entry.grab_focus)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version("Soup", "2.4")
from gi.repository import GLib, Soup

from threading import Thread
from urllib.parse import urlparse
from time import time, sleep

from lollypop.define import App
from lollypop.logger import Logger


class TaskHelper:
    """
        Simple helper for running a task in background
    """

    def __init__(self):
        """
            Init helper
        """
        self.__ratelimit = {}
        self.__retries = {}

    def run(self, command, *args, **kwargs):
        """
            Run command with params and return to callback
            @param command as function
            @param *args as command arguments
            @param **kwargs
            @return thread as Thread
        """
        thread = Thread(target=self.__run,
                        args=(command, kwargs, *args))
        thread.daemon = True
        thread.start()
        return thread

    def load_uri_content(self, uri, cancellable, callback, *args):
        """
            Load uri content async
            @param uri as str
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @callback (uri as str, status as bool, content as bytes, args)
        """
        self.load_uri_content_with_headers(uri, [], cancellable,
                                           callback, *args)

    def load_uri_content_with_headers(self, uri, headers, cancellable,
                                      callback, *args):
        """
            Load uri content async with headers
            @param uri as str
            @param headers as []
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @callback (uri as str, status as bool, content as bytes, args)
        """
        if cancellable is not None and cancellable.is_cancelled():
            callback(uri, False, b"", *args)
        try:
            delay = self.__get_delay_for_uri(uri)
            if delay > 0:
                GLib.timeout_add_seconds(
                                 delay,
                                 self.load_uri_content_with_headers,
                                 uri, headers, cancellable,
                                 callback, *args)
                return

            session = Soup.Session.new()
            session.set_property('accept-language-auto', True)
            session.set_property(
                "user-agent",
                "Lollypop/%s (cedric.bellegarde@adishatz.org)" % App().version)
            msg = Soup.Message.new("GET", uri)
            if headers:
                headers = msg.get_property("request-headers")
                for header in headers:
                    headers.append(header[0],
                                   header[1])
            session.send_async(msg, cancellable,
                               self.__on_load_uri_content, msg, headers,
                               callback, cancellable, uri, *args)
        except Exception as e:
            Logger.warning(
                "HelperTask::load_uri_content_with_headers(): %s" % e)
            callback(uri, False, b"", *args)

    def load_uri_content_sync(self, uri, cancellable=None):
        """
            Load uri
            @param uri as str
            @param cancellable as Gio.Cancellable
            @return (loaded as bool, content as bytes)
        """
        return self.load_uri_content_sync_with_headers(uri, [], cancellable)

    def load_uri_content_sync_with_headers(self, uri, headers,
                                           cancellable=None):
        """
            Load uri
            @param uri as str
            @param headers as []
            @param cancellable as Gio.Cancellable
            @return (loaded as bool, content as bytes)
        """
        try:
            delay = self.__get_delay_for_uri(uri)
            if delay > 0:
                sleep(delay)
                if cancellable is not None and cancellable.is_cancelled():
                    return (False, b"")

            session = Soup.Session.new()
            session.set_property('accept-language-auto', True)
            session.set_property(
                "user-agent",
                "Lollypop/%s (cedric.bellegarde@adishatz.org)" % App().version)
            msg = Soup.Message.new("GET", uri)
            if headers:
                request_headers = msg.get_property("request-headers")
                for header in headers:
                    request_headers.append(header[0], header[1])
            session.send_message(msg)
            response_headers = msg.get_property("response-headers")
            wait = self.__handle_ratelimit(response_headers, uri)
            if wait is None:
                if uri in self.__retries.keys():
                    del self.__retries[uri]
                body = msg.get_property("response-body")
                bytes = body.flatten().get_data()
                return (True, bytes)
            else:
                retries = self.__get_retries_for_uri(uri)
                if retries < 5:
                    parsed = urlparse(uri)
                    self.__ratelimit[parsed.netloc] = wait
                    return self.load_uri_content_sync_with_headers(
                        uri, headers, cancellable)
                else:
                    del self.__retries[uri]
        except Exception as e:
            Logger.warning(
                "TaskHelper::load_uri_content_sync_with_headers(): %s" % e)
            return (False, b"")

    def send_message(self, message, cancellable, callback, *args):
        """
            Send message async
            @param message as Soup.Message
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @callback (uri as str, status as bool, content as bytes, args)
        """
        try:
            uri = message.get_uri().to_string(False)
            delay = self.__get_delay_for_uri(uri)
            if delay > 0:
                GLib.timeout_add_seconds(delay,
                                         self.send_message,
                                         message, cancellable,
                                         callback, *args)
                return

            session = Soup.Session.new()
            session.send_async(message,
                               cancellable,
                               self.__on_message_send_async,
                               message,
                               callback,
                               cancellable,
                               uri,
                               *args)
        except Exception as e:
            Logger.warning("TaskHelper::send_message(): %s" % e)

    def send_message_sync(self, message, cancellable):
        """
            Send message sync
            @param message as Soup.Message
            @param cancellable as Gio.Cancellable
            @return bytes
        """
        try:
            uri = message.get_uri().to_string(False)
            delay = self.__get_delay_for_uri(uri)
            if delay > 0:
                sleep(delay)
                if cancellable is not None and cancellable.is_cancelled():
                    return None

            session = Soup.Session.new()
            stream = session.send(message, cancellable)
            response_headers = message.get_property("response-headers")
            wait = self.__handle_ratelimit(response_headers, uri)
            if wait is None:
                bytes = bytearray(0)
                buf = stream.read_bytes(1024, cancellable).get_data()
                while buf:
                    bytes += buf
                    buf = stream.read_bytes(1024, cancellable).get_data()
                stream.close()
                return bytes
            else:
                retries = self.__get_retries_for_uri(uri)
                if retries < 5:
                    parsed = urlparse(uri)
                    self.__ratelimit[parsed.netloc] = wait
                    return self.send_message_sync(message, cancellable)
                else:
                    del self.__retries[uri]
        except Exception as e:
            Logger.warning("TaskHelper::send_message_sync(): %s" % e)
        return None

#######################
# PRIVATE             #
#######################
    def __get_delay_for_uri(self, uri):
        """
            Get delay for last ratelimit
            @param uri as str
            @return int
        """
        delay = 0
        now = time()
        parsed = urlparse(uri)
        if parsed.netloc in self.__ratelimit.keys():
            wait = self.__ratelimit[parsed.netloc]
            delay = wait - now
            if delay < 0:
                del self.__ratelimit[parsed.netloc]
        return delay

    def __get_retries_for_uri(self, uri):
        """
            Get retries for uri
            @param uri as str
            @return int
        """
        retries = 0
        if uri in self.__retries.keys():
            retries = self.__retries[uri]
        else:
            self.__retries[uri] = 0
        return retries

    def __handle_ratelimit(self, response, uri):
        """
            Set rate limit from response
            @param response as Soup.MessageHeaders
            @param uri as str
            @return next_time as int
        """
        remaining_keys = ["X-RateLimit-Remaining", "X-Rate-Limit-Remaining"]
        reset_keys = ["X-RateLimit-Reset", "X-Rate-Limit-Reset",
                      "X-RateLimit-Reset-In", "X-RateLimit-Reset-At"]
        for key in remaining_keys:
            remaining = response.get(key)
            if remaining is not None:
                break
        for key in reset_keys:
            reset = response.get(key)
            if reset is not None:
                break
        if remaining is None or reset is None:
            return None
        # No more request available
        if (int(remaining) < 1):
            Logger.info(uri)
            Logger.info("X-RateLimit-Remaining: %s" % remaining)
            Logger.info("X-RateLimit-Reset: %s" % reset)
            return int(reset)

        if uri in self.__retries.keys():
            del self.__retries[uri]
        return None

    def __run(self, command, kwd, *args):
        """
            Pass command result to callback
            @param command as function
            @param *args as command arguments
            @param kwd as { "callback": (function, *args) }
        """
        try:
            result = command(*args)
            if "callback" in kwd.keys():
                (callback, *callback_args) = kwd["callback"]
                if callback is not None:
                    GLib.idle_add(callback, result, *callback_args)
        except Exception as e:
            Logger.warning("TaskHelper::__run(): %s: %s -> %s" %
                           (e, command, kwd))

    def __on_read_bytes_async(self, stream, result, content,
                              cancellable, callback, uri, *args):
        """
            Read data from stream, when finished, pass to callback
            @param stream as Gio.InputStream
            @param result as Gio.AsyncResult
            @param cancellable as Gio.Cancellable
            @param content as bytes
            @param callback as function
            @param uri as str
        """
        try:
            content_result = stream.read_bytes_finish(result)
            content_bytes = content_result.get_data()
            if content_bytes:
                content += content_bytes
                stream.read_bytes_async(4096, GLib.PRIORITY_LOW,
                                        cancellable,
                                        self.__on_read_bytes_async,
                                        content, cancellable, callback,
                                        uri, *args)
            else:
                callback(uri, True, bytes(content), *args)
        except Exception as e:
            Logger.warning("TaskHelper::__on_read_bytes_async(): %s" % e)
            callback(uri, False, b"", *args)

    def __on_request_send_async(self, source, result, callback,
                                cancellable, uri, *args):
        """
            Get stream and start reading from it
            @param source as Soup.Session
            @param result as Gio.AsyncResult
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @param uri as str
        """
        try:
            stream = source.send_finish(result)
            # We use a bytearray here as seems that bytes += is really slow
            stream.read_bytes_async(4096, GLib.PRIORITY_LOW,
                                    cancellable, self.__on_read_bytes_async,
                                    bytearray(0), cancellable, callback, uri,
                                    *args)
        except Exception as e:
            Logger.warning("TaskHelper::__on_soup_msg_finished(): %s" % e)
            callback(uri, False, b"", *args)

    def __on_message_send_async(self, source, result, message, callback,
                                cancellable, uri, *args):
        """
            Get stream and start reading from it
            @param source as Soup.Session
            @param result as Gio.AsyncResult
            @param message as Soup.Message
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @param uri as str
        """
        try:
            response_headers = message.get_property("response-headers")
            wait = self.__handle_ratelimit(response_headers, uri)
            if wait is None:
                stream = source.send_finish(result)
                # We use a bytearray here as seems that bytes += is really slow
                stream.read_bytes_async(4096, GLib.PRIORITY_LOW,
                                        cancellable,
                                        self.__on_read_bytes_async,
                                        bytearray(0), cancellable, callback,
                                        uri, *args)
            else:
                parsed = urlparse(uri)
                self.__ratelimit[parsed.netloc] = wait
                retries = self.__get_retries_for_uri(uri)
                if retries < 5:
                    self.__retries[uri] += 1
                    self.send_message(message, cancellable, callback, *args)
                else:
                    del self.__retries[uri]
        except Exception as e:
            Logger.warning("TaskHelper::__on_soup_msg_finished(): %s" % e)
            callback(uri, False, b"", *args)

    def __on_load_uri_content(self, source, result, msg, headers, callback,
                              cancellable, uri, *args):
        """
            Get stream and start reading from it
            @param source as Soup.Session
            @param result as Gio.AsyncResult
            @param msg as Soup.Message
            @param headers as []
            @param cancellable as Gio.Cancellable
            @param callback as a function
            @param uri as str
        """
        try:
            response_headers = msg.get_property("response-headers")
            wait = self.__handle_ratelimit(response_headers, uri)
            if wait is None:
                stream = source.send_finish(result)
                # We use a bytearray here as seems that bytes += is really slow
                stream.read_bytes_async(4096, GLib.PRIORITY_LOW,
                                        cancellable,
                                        self.__on_read_bytes_async,
                                        bytearray(0), cancellable, callback,
                                        uri, *args)
            else:
                parsed = urlparse(uri)
                self.__ratelimit[parsed.netloc] = wait
                retries = self.__get_retries_for_uri(uri)
                if retries < 5:
                    self.__retries[uri] += 1
                    self.load_uri_content_sync_with_headers(uri, headers,
                                                            callback, *args)
                else:
                    del self.__retries[uri]
        except Exception as e:
            Logger.warning("TaskHelper::__on_soup_msg_finished(): %s" % e)
            callback(uri, False, b"", *args)



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json

from lollypop.logger import Logger
from lollypop.utils import emit_signal, get_network_available
from lollypop.helper_web_musicbrainz import MusicBrainzWebHelper
from lollypop.helper_web_save import SaveWebHelper
from lollypop.define import App


class MusicBrainzSearch(SaveWebHelper, MusicBrainzWebHelper):
    """
        Search for MusicBrainz
    """

    def __init__(self):
        """
            Init object
        """
        SaveWebHelper.__init__(self)
        MusicBrainzWebHelper.__init__(self)

    def get(self, search, storage_type, cancellable):
        """
            Get albums for search
            We need a thread because we are going to populate DB
            @param search as str
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        if not get_network_available("MUSICBRAINZ"):
            emit_signal(self, "finished")
            return
        try:
            uri = "http://musicbrainz.org/ws/2/release/?fmt=json&query=%s" %\
                search
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for release in decode["releases"]:
                    payload = self.lollypop_album_payload(release)
                    self.save_album_payload_to_db(payload,
                                                  storage_type,
                                                  True,
                                                  cancellable)
        except Exception as e:
            Logger.warning("MusicBrainzSearch::get(): %s", e)
        if not cancellable.is_cancelled():
            emit_signal(self, "finished")

    def load_tracks(self, album, cancellable):
        """
            Load tracks for album
            @param album as Album
            @param cancellable as Gio.Cancellable
        """
        try:
            mbid = album.uri.replace("mb:", "")
            tracks = self.get_tracks_payload(mbid, cancellable)
            # We want to share the same item as lp_album_id may change
            album_item = album.collection_item
            for track in tracks:
                lollypop_payload = self.lollypop_track_payload(track)
                self.save_track_payload_to_db(lollypop_payload,
                                              album_item,
                                              album.storage_type,
                                              False,
                                              cancellable)
        except Exception as e:
            Logger.error("MusicBrainzSearch::load_tracks(): %s", e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.define import App
from lollypop.utils import get_network_available
from lollypop.similars_local import LocalSimilars
from lollypop.similars_spotify import SpotifySimilars
from lollypop.similars_lastfm import LastFMSimilars
from lollypop.similars_deezer import DeezerSimilars


class Similars():
    """
        Search similar artists
    """
    def __init__(self):
        """
            Init similars
        """
        self.__local_helper = LocalSimilars()
        self.__spotify_helper = SpotifySimilars()
        self.__lastfm_helper = LastFMSimilars()
        self.__deezer_helper = DeezerSimilars()

    def get_similar_artists(self, artist_ids, cancellable):
        """
            Get similar artists
            @param artist_ids as [int]
            @param cancellable as Gio.Cancellable
            @return [(str, None)]
        """
        artist_names = []
        result = []
        for artist_id in artist_ids:
            artist_names.append(App().artists.get_name(artist_id))

        if get_network_available("DEEZER"):
            result = self.__deezer_helper.get_similar_artists(
                artist_names, cancellable)
        if not result and get_network_available("SPOTIFY"):
            result = self.__spotify_helper.get_similar_artists(
                artist_names, cancellable)
        if not result and get_network_available("LASTFM"):
            result = self.__lastfm_helper.get_similar_artists(
                artist_names, cancellable)
        if not result:
            result = self.__local_helper.get_similar_artists(
                artist_names, cancellable)
        return result



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.logger import Logger


class FrameTag:
    """
        Bytes representing a frame
    """

    def __init__(self, bytes):
        """
            Init tag reader
            @param bytes as bytes
        """
        try:
            self.__key = bytes[0:4].decode("utf-8")
        except Exception as e:
            Logger.error("FrameTag::__init__(self): %s" % e)
            self.__key = "None"
        self.__bytes = bytes

    @property
    def frame(self):
        """
            Get frame
            @return bytes
        """
        return self.__bytes[10:]

    @property
    def encoding(self):
        """
            Get frame
            @return bytes
        """
        return self.frame[0:1]

    @property
    def key(self):
        """
            Get frame key
            @return str
        """
        return self.__key

    @property
    def string(self):
        """
            String representation of data
            @return str
        """
        return ""



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _

from lollypop.define import App, StorageType
from lollypop.utils import get_network_available


class SuggestionsMenu(Gio.Menu):
    """
        Contextual menu for suggestions
    """
    def __init__(self, header=False):
        """
            Init search menu
            @param header as bool
        """
        Gio.Menu.__init__(self)
        section = Gio.Menu()
        if header:
            from lollypop.menu_header import MenuHeader
            self.append_item(MenuHeader(
                                 _("Suggestions"),
                                 "org.gnome.Lollypop-suggestions-symbolic"))
        mask = App().settings.get_value("suggestions-mask").get_int32()
        action = Gio.SimpleAction.new_stateful(
            "spotify_new_releases",
            None,
            GLib.Variant("b", mask & StorageType.SPOTIFY_NEW_RELEASES))
        action.connect("change-state",
                       self.__on_change_state,
                       StorageType.SPOTIFY_NEW_RELEASES)
        action.set_enabled(get_network_available("SPOTIFY"))
        App().add_action(action)
        menu_item = Gio.MenuItem.new(_("New releases on Spotify"),
                                     "app.spotify_new_releases")
        section.append_item(menu_item)
        action = Gio.SimpleAction.new_stateful(
            "spotify_similars",
            None,
            GLib.Variant("b", mask & StorageType.SPOTIFY_SIMILARS))
        action.connect("change-state",
                       self.__on_change_state,
                       StorageType.SPOTIFY_SIMILARS)
        action.set_enabled(get_network_available("SPOTIFY"))
        App().add_action(action)
        menu_item = Gio.MenuItem.new(_("Suggestions from Spotify"),
                                     "app.spotify_similars")
        section.append_item(menu_item)
        action = Gio.SimpleAction.new_stateful(
            "deezer_charts",
            None,
            GLib.Variant("b", mask & StorageType.DEEZER_CHARTS))
        action.connect("change-state",
                       self.__on_change_state,
                       StorageType.DEEZER_CHARTS)
        action.set_enabled(get_network_available("DEEZER"))
        App().add_action(action)
        menu_item = Gio.MenuItem.new(_("Top albums on Deezer"),
                                     "app.deezer_charts")
        section.append_item(menu_item)
        self.append_section(_("From the Web"), section)

#######################
# PRIVATE             #
#######################
    def __clean_storage_type(self, state, mask):
        """
            Clear storage type if state false
            @param state as bool
            @param mask as int
        """
        if state:
            return
        album_ids = App().albums.get_for_storage_type(mask)
        for album_id in album_ids:
            # EPHEMERAL with not tracks will be cleaned below
            App().albums.set_storage_type(album_id,
                                          StorageType.EPHEMERAL)
            App().tracks.remove_album(album_id)
        App().tracks.clean()
        App().albums.clean()
        App().artists.clean()

    def __handle_mask_change(self, state, mask):
        """
            Update mask value in settings
            @param state as bool
            @param mask as int
        """
        if App().ws_director.collection_ws is not None:
            if not App().ws_director.collection_ws.stop():
                GLib.timeout_add(500, self.__handle_mask_change, state, mask)
                return
        suggestion_mask = App().settings.get_value(
            "suggestions-mask").get_int32()
        if state:
            suggestion_mask |= mask
        else:
            suggestion_mask &= ~mask
        App().settings.set_value("suggestions-mask",
                                 GLib.Variant("i", suggestion_mask))
        self.__clean_storage_type(state, mask)
        if App().ws_director.collection_ws is not None:
            App().ws_director.collection_ws.start()

    def __on_change_state(self, action, state, mask):
        """
            Update settings value
            @param action as Gio.SimpleAction
            @param state as bool
            @param mask as int
        """
        action.set_state(state)
        self.__handle_mask_change(state, mask)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from pickle import dump, load

from lollypop.logger import Logger
from lollypop.define import LOLLYPOP_DATA_PATH


class HistoryContainer:
    """
        Navigation history
        Offload old items and reload them on the fly
    """

    __MAX_HISTORY_ITEMS = 4

    def __init__(self):
        """
            Init history
        """
        self.__items = []

    def add_view(self, view):
        """
            Add view to history
            @param view as View
        """
        if view.args is None:
            return
        view_class = view.__class__
        self.__items.append((view, view_class, view.args, view.sidebar_id,
                             view.selection_ids, view.position))
        # Offload history if too many items
        if self.count >= self.__MAX_HISTORY_ITEMS:
            (view, _class, args, sidebar_id,
             selection_ids, position) = self.__items[-self.__MAX_HISTORY_ITEMS]
            if view is not None:
                view.destroy()
                # This view can't be offloaded
                if args is None:
                    del self.__items[-self.__MAX_HISTORY_ITEMS]
                else:
                    self.__items[-self.__MAX_HISTORY_ITEMS] =\
                        (None, _class, args, sidebar_id,
                         selection_ids, position)

    def pop(self, index=-1):
        """
            Pop last view from history
            @param index as int
            @return View
        """
        try:
            if not self.__items:
                return None
            (view, _class, args, sidebar_id,
             selection_ids, position) = self.__items.pop(index)
            # View is offloaded, create a new one
            if view is None:
                view = self.__get_view_from_class(_class, args)
                view.set_sidebar_id(sidebar_id)
                view.set_selection_ids(selection_ids)
                view.set_populated_scrolled_position(position)
            return view
        except Exception as e:
            Logger.error("AdaptiveHistory::pop(): %s" % e)
            self.__items = []

    def remove(self, view):
        """
            Remove view from history
            @param view as View
        """
        for (_view, _class, args, sidebar_id,
             selection_ids, position) in self.__items:
            if _view == view:
                self.__items.remove((_view, _class, args, sidebar_id,
                                     selection_ids, position))
                break

    def clear(self):
        """
            Clear history
        """
        for (view, _class, args, sidebar_id,
             selection_ids, position) in self.__items:
            if view is not None:
                view.stop()
                view.destroy_later()
        self.__items = []

    def save(self):
        """
            Save history
        """
        try:
            history = []
            for (_view, _class, args, sidebar_id,
                 selection_ids, position) in self.__items[-50:]:
                history.append((None, _class, args, sidebar_id,
                                selection_ids, position))
            with open(LOLLYPOP_DATA_PATH + "/history.bin", "wb") as f:
                dump(history, f)
        except Exception as e:
            Logger.error("AdaptiveHistory::save(): %s" % e)

    def load(self):
        """
            Load history
        """
        try:
            self.__items = load(
                open(LOLLYPOP_DATA_PATH + "/history.bin", "rb"))
        except Exception as e:
            Logger.error("AdaptiveHistory::load(): %s" % e)

    def exists(self, view):
        """
            True if view exists in history
            @return bool
        """
        for (_view, _class, args, sidebar_id,
             selection_ids, position) in self.__items:
            if _view == view:
                return True
        return False

    @property
    def items(self):
        """
            Get history items
            @return [(View, class, {})]
        """
        return self.__items

    @property
    def count(self):
        """
            Get history item count
            @return int
        """
        return len(self.__items)

############
# PRIVATE  #
############
    def __get_view_from_class(self, _class, args):
        """
            Get view from history
            @param _class as class
            @param args as {}
            @return View
        """
        try:
            view = _class(**args)
            # Start populating the view
            if hasattr(view, "populate"):
                view.populate()
            view.show()
            return view
        except Exception as e:
            Logger.warning(
                "AdaptiveHistory::__get_view_from_class(): %s, %s",
                _class, e)
        return None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk, Gdk, GdkPixbuf, Gio

import cairo
from random import shuffle

from lollypop.define import App, Type
from lollypop.objects_album import Album
from lollypop.utils import get_round_surface, emit_signal, get_icon_name
from lollypop.widgets_flowbox_rounded import RoundedFlowBoxWidget


class RoundedAlbumsWidget(RoundedFlowBoxWidget):
    """
        Rounded widget showing cover for up to 9 albums
    """
    _ALBUMS_COUNT = 10

    def __init__(self, data, name, sortname, view_type, font_height):
        """
            Init widget
            @param data as object
            @param name as str
            @param sortname as str
            @param view_type as ViewType
            @param font_height as int
        """
        RoundedFlowBoxWidget.__init__(self, data, name, sortname,
                                      view_type, font_height)
        self._genre = Type.NONE
        self.__album_ids = []
        self.__cancellable = Gio.Cancellable()
        self._scale_factor = self.get_scale_factor()
        self.connect("unmap", self.__on_unmap)

    def populate(self):
        """
            Populate widget content
        """
        RoundedFlowBoxWidget.populate(self)
        self._artwork.get_style_context().add_class("rounded-icon-large")

    def set_artwork(self):
        """
            Set artwork
        """
        RoundedFlowBoxWidget.set_artwork(self)
        if App().art.exists_in_cache(self.artwork_name,
                                     "ROUNDED",
                                     self._art_size,
                                     self._art_size):
            App().task_helper.run(
                App().art.get_from_cache,
                self.artwork_name, "ROUNDED",
                self._art_size, self._art_size,
                callback=(self.__on_load_from_cache,))
        else:
            self.__album_ids = self._get_album_ids()
            shuffle(self.__album_ids)
            App().task_helper.run(self._create_surface, True)

#######################
# PROTECTED           #
#######################
    def _create_surface(self, set_surface):
        """
            Get artwork surface
            @param set_surface as bool
            @return cairo.Surface
        """
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                     self._art_size,
                                     self._art_size)
        ctx = cairo.Context(surface)
        ctx.rectangle(0, 0, self._art_size, self._art_size)
        ctx.set_source_rgb(1, 1, 1)
        ctx.fill()
        album_ids = list(self.__album_ids)
        album_pixbufs = []
        album_scaled_pixbufs = []
        while album_ids and len(album_pixbufs) != 9:
            album_id = album_ids.pop(0)
            pixbuf = App().album_art.get(Album(album_id),
                                         self._art_size,
                                         self._art_size,
                                         self._scale_factor)
            if pixbuf is not None:
                album_pixbufs.append(pixbuf)
        if len(album_pixbufs) == 0:
            self.__cover_size = self._art_size / 2
            positions = [(0.5, 0.5)]
        elif 1 <= len(album_pixbufs) <= 2:
            self.__cover_size = self._art_size
            positions = [(0, 0)]
        elif 3 <= len(album_pixbufs) <= 5:
            self.__cover_size = self._art_size / 2
            positions = [(0, 0), (1, 0),
                         (0, 1), (1, 1)]
        else:
            self.__cover_size = self._art_size / 3
            positions = [(0, 0), (1, 0), (2, 0),
                         (0, 1), (1, 1), (2, 1),
                         (0, 2), (1, 2), (2, 2)]
        while album_pixbufs:
            pixbuf = album_pixbufs.pop(0)
            newpix = pixbuf.scale_simple(self.__cover_size,
                                         self.__cover_size,
                                         GdkPixbuf.InterpType.NEAREST)
            album_scaled_pixbufs.append(newpix)

        if len(album_scaled_pixbufs) == 0:
            theme = Gtk.IconTheme.get_default()
            category_icon = get_icon_name(self._genre)
            symbolic = theme.lookup_icon(category_icon,
                                         self.__cover_size,
                                         Gtk.IconLookupFlags.USE_BUILTIN)
            if symbolic is not None:
                pixbuf = symbolic.load_icon()
                album_scaled_pixbufs.append(pixbuf)

        if len(album_scaled_pixbufs) == 3:
            album_scaled_pixbufs.append(album_scaled_pixbufs[0])
        if len(album_scaled_pixbufs) == 6:
            album_scaled_pixbufs.append(album_scaled_pixbufs[2])
            album_scaled_pixbufs.append(album_scaled_pixbufs[1])
            album_scaled_pixbufs.append(album_scaled_pixbufs[0])
        if len(album_scaled_pixbufs) == 7:
            album_scaled_pixbufs.append(album_scaled_pixbufs[1])
            album_scaled_pixbufs.append(album_scaled_pixbufs[0])
        if len(album_scaled_pixbufs) == 8:
            album_scaled_pixbufs.append(album_scaled_pixbufs[0])

        self.__draw_surface(surface, ctx, positions,
                            album_scaled_pixbufs, set_surface)

#######################
# PRIVATE             #
#######################
    def __set_surface(self, surface):
        """
            Set artwork from surface
            @param surface as cairo.Surface
        """
        if self.__cancellable.is_cancelled():
            return
        rounded = get_round_surface(
            surface, self._scale_factor, self._art_size / 4)
        self._artwork.set_from_surface(rounded)
        App().art.add_to_cache(self.artwork_name,
                               rounded,
                               "ROUNDED",
                               self._artwork.get_scale_factor())
        emit_signal(self, "populated")

    def __draw_surface(self, surface, ctx, positions,
                       album_pixbufs, set_surface):
        """
            Draw surface for first available album
            @param surface as cairo.Surface
            @param ctx as Cairo.context
            @param positions as {}
            @param album_pixbufs as []
            @param set_surface as bool
            @thread safe
        """
        # Workaround Gdk not being thread safe
        def draw_pixbuf(surface, ctx, pixbuf, positions, album_pixbufs):
            if self.__cancellable.is_cancelled():
                return
            (x, y) = positions.pop(0)
            x *= self.__cover_size
            y *= self.__cover_size
            subsurface = Gdk.cairo_surface_create_from_pixbuf(
                pixbuf, self._scale_factor, None)
            ctx.translate(x, y)
            ctx.set_source_surface(subsurface, 0, 0)
            ctx.paint()
            ctx.translate(-x, -y)
            self.__draw_surface(surface, ctx, positions,
                                album_pixbufs, set_surface)
        if self.__cancellable.is_cancelled():
            return
        elif album_pixbufs and len(positions) > 0:
            pixbuf = album_pixbufs.pop(0)
            if pixbuf is None:
                GLib.idle_add(self.__draw_surface, surface,
                              ctx, positions, album_pixbufs, set_surface)
            else:
                GLib.idle_add(draw_pixbuf, surface,
                              ctx, pixbuf, positions, album_pixbufs)
        else:
            GLib.idle_add(self.__set_surface, surface)

    def __on_load_from_cache(self, pixbuf):
        """
            Set artwork surface
            @param pixbuf as GdkPixbuf.Pixbuf
        """
        if not self.__cancellable.is_cancelled() and pixbuf is not None:
            surface = Gdk.cairo_surface_create_from_pixbuf(
                pixbuf, self._artwork.get_scale_factor(), None)
            self._artwork.set_from_surface(
                    get_round_surface(surface, self._scale_factor,
                                      self._art_size / 4))
        emit_signal(self, "populated")

    def __on_unmap(self, widget):
        """
            Cancel drawing
            @param widget as Gtk.Widget
        """
        self.__cancellable.cancel()
        self.__cancellable = Gio.Cancellable()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gio

from gettext import gettext as _

from lollypop.define import App, StorageType
from lollypop.define import ViewType, MARGIN
from lollypop.search import Search
from lollypop.view import View
from lollypop.utils import sql_escape
from lollypop.objects_album import Album
from lollypop.objects_track import Track
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.view_artists_line import ArtistsSearchLineView
from lollypop.view_albums_line import AlbumsSearchLineView
from lollypop.view_tracks_search import SearchTracksView
from lollypop.widgets_banner_search import SearchBannerWidget


class SearchGrid(Gtk.Grid):
    """
        A grid for search
    """

    def __init__(self, storage_type):
        """
            Init grid
            @param storage_type as StorageType
        """
        Gtk.Grid.__init__(self)
        self.set_row_spacing(MARGIN)
        self.get_style_context().add_class("padding")
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_property("valign", Gtk.Align.START)
        self.__artists_line_view = ArtistsSearchLineView(storage_type)
        self.__albums_line_view = AlbumsSearchLineView(storage_type)
        self.__search_tracks_view = SearchTracksView()
        self.add(self.__albums_line_view)
        self.add(self.__artists_line_view)
        self.add(self.__search_tracks_view)

    @property
    def search_tracks_view(self):
        """
            Get SearchTracksView
            @return SearchTracksView
        """
        return self.__search_tracks_view

    @property
    def artists_line_view(self):
        """
            Get ArtistsSearchLineView
            @return ArtistsSearchLineView
        """
        return self.__artists_line_view

    @property
    def albums_line_view(self):
        """
            Get AlbumsSearchLineView
            @return AlbumsSearchLineView
        """
        return self.__albums_line_view


class SearchStack(Gtk.Stack):
    """
        A stack for search
    """

    def __init__(self, storage_type):
        """
            Init stack
            @param storage_type as StorageType
        """
        Gtk.Stack.__init__(self)
        self.get_style_context().add_class("padding")
        self.__current_child = None
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.set_transition_duration(100)
        for i in range(0, 2):
            grid = SearchGrid(storage_type)
            grid.show()
            self.add(grid)

    def new_current_child(self):
        """
            Set a new current child for search
            This is not visible child
        """
        if self.__current_child is not None:
            self.__current_child.search_tracks_view.clear()
            self.__current_child.artists_line_view.clear()
            self.__current_child.albums_line_view.clear()
        for child in self.get_children():
            if child != self.get_visible_child():
                self.__current_child = child
                break

    @property
    def current_child(self):
        """
            Get non visible child
            @return SearchGrid
        """
        if self.__current_child is None:
            self.new_current_child()
        return self.__current_child


class SearchView(View, Gtk.Bin, SignalsHelper):
    """
        View for searching albums/tracks
    """

    @signals_map
    def __init__(self, initial_search, view_type):
        """
            Init Popover
            @param initial_search as str
        """
        View.__init__(self,
                      StorageType.COLLECTION |
                      StorageType.SAVED |
                      StorageType.SEARCH |
                      StorageType.EPHEMERAL |
                      StorageType.SPOTIFY_NEW_RELEASES |
                      StorageType.SPOTIFY_SIMILARS,
                      ViewType.SEARCH |
                      ViewType.SCROLLED |
                      ViewType.OVERLAY)
        Gtk.Bin.__init__(self)
        self.__timeout_id = None
        self.__current_search = ""
        self.__search = Search()
        self.__search.set_web_search(
            App().settings.get_value("web-search").get_string())
        self.__cancellable = Gio.Cancellable()
        self._empty_message = _("Search for artists, albums and tracks")
        self._empty_icon_name = "edit-find-symbolic"
        self.__cancellable = Gio.Cancellable()
        self.__banner = SearchBannerWidget()
        self.__banner.show()
        self.__stack = SearchStack(self.storage_type)
        self.__stack.show()
        self.add_widget(self.__stack, self.__banner)
        self.__banner.entry.connect("changed", self._on_search_changed)
        self.show_placeholder(True,
                              _("Search for artists, albums and tracks"))
        self.set_search(initial_search)
        return [
                (self.__search, "match-artist", "_on_match_artist"),
                (self.__search, "match-album", "_on_match_album"),
                (self.__search, "match-track", "_on_match_track"),
                (self.__search, "finished", "_on_search_finished"),
                (App().settings, "changed::web-search",
                 "_on_web_search_changed")
        ]

    def populate(self):
        """
            Populate search
            in db based on text entry current text
        """
        self.cancel()
        self.__stack.new_current_child()
        if len(self.__current_search) > 1:
            self.__banner.spinner.start()
            current_search = self.__current_search.lower()
            self.__search.get(current_search, self.__cancellable)
        else:
            self.show_placeholder(True,
                                  _("Search for artists, albums and tracks"))
            self.__banner.spinner.stop()

    def set_search(self, search):
        """
            Set search text
            @param search as str
        """
        self.__banner.entry.set_text(search)
        self.__banner.entry.grab_focus()

    def grab_focus(self):
        """
            Make search entry grab focus
        """
        self.__banner.entry.grab_focus()

    def cancel(self):
        """
            Cancel current search and replace cancellable
        """
        self.__cancellable.cancel()
        self.__cancellable = Gio.Cancellable()

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        search = self.__banner.entry.get_text().strip()
        return {"initial_search": search,
                "view_type": self.view_type}

#######################
# PROTECTED           #
#######################
    def _on_map(self, widget):
        """
            Disable shortcuts and update buttons
            @param widget as Gtk.Widget
        """
        View._on_map(self, widget)
        App().enable_special_shortcuts(False)

    def _on_unmap(self, widget):
        """
            Cancel current loading and enable shortcuts
            @param widget as Gtk.Widget
        """
        View._on_unmap(self, widget)
        App().enable_special_shortcuts(True)
        self.cancel()
        self.__banner.spinner.stop()

    def _on_match_artist(self, search, artist_id, storage_type):
        """
            Add a new artist to view
            @param search as *Search
            @param artist_id as int
            @param storage_type as StorageType
        """
        if storage_type & StorageType.SEARCH:
            self.__stack.current_child.artists_line_view.show()
            self.__stack.current_child.artists_line_view.add_value(artist_id)
            self.show_placeholder(False)

    def _on_match_album(self, search, album_id, storage_type):
        """
            Add a new album to view
            @param search as *Search
            @param artist_id as int
            @param storage_type as StorageType
        """
        if storage_type & StorageType.SEARCH:
            artist_match = False
            album = Album(album_id)
            if album.artists:
                artist = sql_escape(album.artists[0])
                search = sql_escape(self.__current_search)
                artist_match = artist.find(search) != -1
            if artist_match:
                self._on_match_artist(search,
                                      album.artist_ids[0],
                                      storage_type)
            else:
                self.__stack.current_child.albums_line_view.show()
                self.__stack.current_child.albums_line_view.add_value(album)
                self.show_placeholder(False)

    def _on_match_track(self, search, track_id, storage_type):
        """
            Add a new track to view
            @param search as *Search
            @param track_id as int
            @param storage_type as StorageType
        """
        if storage_type & StorageType.SEARCH:
            track = Track(track_id)
            self.__stack.current_child.search_tracks_view.show()
            self.__stack.current_child.search_tracks_view.append_row(track)
            self.show_placeholder(False)

    def _on_search_finished(self, search_handler, last):
        """
            Stop spinner and show placeholder if not result
            @param search_handler as LocalSearch/WebSearch
            @param last as bool
        """
        tracks_len = len(
            self.__stack.current_child.search_tracks_view.children)
        albums_len = len(
            self.__stack.current_child.albums_line_view.children)
        artists_len = len(
            self.__stack.current_child.artists_line_view.children)
        empty = albums_len == 0 and tracks_len == 0 and artists_len == 0
        self.__stack.set_visible_child(self.__stack.current_child)
        if last:
            self.__banner.spinner.stop()
            if empty:
                self.show_placeholder(True, _("No results for this search"))

    def _on_web_search_changed(self, settings, value):
        """
            Show/hide labels
            @param settings as Gio.Settings
            @param value as GLib.Variant
        """
        self.__search.set_web_search(
            App().settings.get_value("web-search").get_string())

#######################
# PRIVATE             #
#######################
    def __set_no_result_placeholder(self):
        """
            Set placeholder for no result
        """
        self.__placeholder.set_text()

    def _on_search_changed(self, widget):
        """
            Timeout filtering
            @param widget as Gtk.TextEntry
        """
        if self.__timeout_id is not None:
            GLib.source_remove(self.__timeout_id)
        self.__timeout_id = GLib.timeout_add(
                500,
                self.__on_search_changed_timeout)

    def __on_search_changed_timeout(self):
        """
            Populate widget
        """
        self.__timeout_id = None
        new_search = self.__banner.entry.get_text().strip()
        if self.__current_search != new_search:
            self.__current_search = new_search
            self.populate()

    def __on_button_clicked(self, button):
        """
            Reload search for current button
            @param button as Gtk.RadioButton
        """
        if button.get_active():
            self.__current_search = self.__banner.entry.get_text().strip()
            if self.__current_search:
                self.populate()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from gettext import gettext as _
import itertools

from lollypop.sqlcursor import SqlCursor
from lollypop.define import App, StorageType, Type, LovedFlags
from lollypop.utils import noaccents, make_subrequest


class TracksDatabase:
    """
        All functions take a sqlite cursor as last parameter,
        set another one if you"re in a thread
    """

    def __init__(self, db):
        """
            Init tracks database object
            @param db as database
        """
        self.__db = db

    def add(self, name, uri, duration, tracknumber, discnumber, discname,
            album_id, year, timestamp, popularity, rate, loved, ltime, mtime,
            mb_track_id, lp_track_id, bpm, storage_type):
        """
            Add a new track to database
            @param name as string
            @param uri as string,
            @param duration as int
            @param tracknumber as int
            @param discnumber as int
            @param discname as str
            @param album_id as int
            @param year as int
            @param timestamp as int
            @param popularity as int
            @param rate as int
            @param loved as bool
            @param ltime as int
            @param mtime as int
            @param mb_track_id as str
            @param lp_track_id as str
            @param bpm as double
            @return inserted rowid as int
            @warning: commit needed
        """
        with SqlCursor(self.__db, True) as sql:
            result = sql.execute(
                "INSERT INTO tracks (name, uri, duration, tracknumber,\
                discnumber, discname, album_id,\
                year, timestamp, popularity, rate, loved,\
                ltime, mtime, mb_track_id, lp_track_id, bpm, storage_type)\
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (name, uri, duration, tracknumber, discnumber,
                 discname, album_id, year, timestamp, popularity,
                 rate, loved, ltime, mtime, mb_track_id, lp_track_id,
                 bpm, storage_type))
            return result.lastrowid

    def add_artist(self, track_id, artist_id):
        """
            Add artist to track
            @param track_id as int
            @param artist_id as int
            @warning: commit needed
        """
        with SqlCursor(self.__db, True) as sql:
            artists = self.get_artist_ids(track_id)
            if artist_id not in artists:
                sql.execute("INSERT INTO "
                            "track_artists (track_id, artist_id)"
                            "VALUES (?, ?)", (track_id, artist_id))

    def add_genre(self, track_id, genre_id):
        """
            Add genre to track
            @param track_id as int
            @param genre_id as int
            @warning: commit needed
        """
        with SqlCursor(self.__db, True) as sql:
            genres = self.get_genre_ids(track_id)
            if genre_id not in genres:
                sql.execute("INSERT INTO\
                             track_genres (track_id, genre_id)\
                             VALUES (?, ?)",
                            (track_id, genre_id))

    def get_ids(self, storage_type, skipped):
        """
            Return all internal track ids
            @param storage_type as StorageType
            @param skipped as bool
            @return track ids as [int]
        """
        with SqlCursor(self.__db) as sql:
            filters = (storage_type,)
            request = "SELECT rowid FROM tracks\
                       WHERE storage_type & ?"
            if not skipped:
                request += " AND not loved &? "
                filters += (LovedFlags.SKIPPED,)
            request += " ORDER BY album_id"
            result = sql.execute(request, filters)
            return list(itertools.chain(*result))

    def get_ids_for_name(self, name):
        """
            Return tracks ids with name
            @param name as str
            @return track id as [int]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM tracks WHERE noaccents(name)=?\
                                  COLLATE NOCASE",
                                 (noaccents(name),))
            return list(itertools.chain(*result))

    def get_id_by_uri(self, uri):
        """
            Return track id for uri
            @param uri as str
            @return track id as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT rowid FROM tracks WHERE uri=?",
                                 (uri,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_id_by_basename_duration(self, basename, duration):
        """
            Get track id by basename
            @param basename as str
            @param duration as int
            @return track_id as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT rowid FROM tracks\
                                  WHERE uri like ? AND duration=?",
                                 ("%" + basename, duration))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_name(self, track_id):
        """
            Get track name for track id
            @param track_id as int
            @return Name as string
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT name FROM tracks WHERE rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_year(self, track_id):
        """
            Get track year
            @param track_id as int
            @return year as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT year FROM tracks WHERE rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v and v[0]:
                return v[0]
            return None

    def get_storage_type(self, track_id):
        """
            Get storage type
            @param track_id as int
            @return int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT storage_type FROM tracks WHERE\
                                 rowid=?", (track_id,))

            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_timestamp(self, track_id):
        """
            Get track timestamp
            @param track_id as int
            @return timestamp as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT timestamp FROM tracks WHERE rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v and v[0]:
                return v[0]
            return None

    def get_timestamp_for_album(self, album_id):
        """
            Get album timestamp based on tracks
            Use most used timestamp by tracks
            @param album_id as int
            @return int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT timestamp,\
                                  COUNT(timestamp) AS occurrence\
                                  FROM tracks\
                                  WHERE tracks.album_id=?\
                                  GROUP BY timestamp\
                                  ORDER BY occurrence DESC\
                                  LIMIT 1", (album_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_rate(self, track_id):
        """
            Get track rate
            @param track_id as int
            @return rate as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT rate FROM tracks WHERE rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v:
                return v[0]
            return 0

    def get_uri(self, track_id):
        """
            Get track uri for track id
            @param track_id as int
            @return uri as string
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT uri FROM tracks WHERE rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def set_uri(self, track_id, uri):
        """
            Set track uri
            @param track_id as int
            @param uri as string
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks SET uri=?\
                         WHERE rowid=?",
                        (uri, track_id))

    def set_storage_type(self, track_id, storage_type):
        """
            Set storage type
            @param track_id as int
            @param storage_type as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks SET storage_type=?\
                         WHERE rowid=?",
                        (storage_type, track_id))

    def set_rate(self, track_id, rate):
        """
            Set track rate
            @param track_id as int
            @param rate as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks SET rate=?\
                         WHERE rowid=?",
                        (rate, track_id))

    def get_album_id(self, track_id):
        """
            Get album id for track id
            @param track_id as int
            @return album id as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT album_id FROM tracks WHERE rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return -1

    def get_mb_track_id(self, track_id):
        """
            Get MusicBrainz recording id for track id
            @param track_id as int
            @return recording id as str
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT mb_track_id FROM tracks\
                                  WHERE rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_id_for_lp_track_id(self, lp_track_id):
        """
            Get track id for Lollypop recording id
            @param Lollypop id as str
            @return track id as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT rowid FROM tracks\
                                  WHERE lp_track_id=?", (lp_track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return -1

    def get_album_name(self, track_id):
        """
            Get album name for track id
            @param track_id as int
            @return album name as str
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT albums.name from albums,tracks\
                                  WHERE tracks.rowid=? AND\
                                  tracks.album_id=albums.rowid", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return _("Unknown")

    def get_artist_ids(self, track_id):
        """
            Get artist ids
            @param track_id as int
            @return artist ids as [int]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT artist_id FROM track_artists\
                                  WHERE track_id=?", (track_id,))
            return list(itertools.chain(*result))

    def get_mb_artist_ids(self, track_id):
        """
            Get MusicBrainz artist ids
            @param track_id as int
            @return artist ids as [int]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT mb_artist_id\
                                  FROM artists, track_artists\
                                  WHERE track_artists.track_id=?\
                                  AND track_artists.artist_id=artists.rowid",
                                 (track_id,))
            return list(itertools.chain(*result))

    def get_artists(self, track_id):
        """
            Get artist names
            @param track_id as int
            @return artists as [str]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT name FROM artists, track_artists\
                                  WHERE track_artists.track_id=?\
                                  AND track_artists.artist_id=artists.rowid",
                                 (track_id,))
            return list(itertools.chain(*result))

    def get_album_genre_ids(self, album_id):
        """
            Get album genre ids based on tracks
            @param album_id as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT track_genres.genre_id\
                                  FROM tracks, track_genres\
                                  WHERE tracks.album_id=? AND\
                                  track_genres.track_id=tracks.rowid",
                                 (album_id,))
            return list(itertools.chain(*result))

    def get_genre_ids(self, track_id):
        """
            Get genre ids
            @param track_id as int
            @return genre ids as [int]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT genre_id FROM track_genres\
                                  WHERE track_id=?", (track_id,))
            return list(itertools.chain(*result))

    def get_genres(self, track_id):
        """
            Get genres
            @param track_id as int
            @return [str]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT name FROM genres, track_genres\
                                  WHERE track_genres.track_id=?\
                                  AND track_genres.genre_id=genres.rowid",
                                 (track_id,))
            return list(itertools.chain(*result))

    def get_mtimes(self):
        """
            Get mtime for tracks
            @return dict of {uri as string: mtime as int}
        """
        with SqlCursor(self.__db) as sql:
            mtimes = {}
            result = sql.execute("SELECT DISTINCT uri, mtime\
                                  FROM tracks WHERE storage_type & ?",
                                 (StorageType.COLLECTION,))
            for row in result:
                mtimes.update((row,))
            return mtimes

    def remove_album(self, album_id, commit=True):
        """
            Remove album
            @param album_id as int
            @param commit as bool
        """
        with SqlCursor(self.__db, commit) as sql:
            sql.execute("DELETE FROM tracks WHERE album_id=?", (album_id,))

    def del_non_persistent(self, commit=True):
        """
            Delete non persistent tracks
            @param commit as bool
        """
        with SqlCursor(self.__db, commit) as sql:
            sql.execute("DELETE FROM tracks WHERE storage_type & ?",
                        (StorageType.EPHEMERAL | StorageType.EXTERNAL,))

    def del_persistent(self, commit=True):
        """
            Delete persistent tracks
            @param commit as bool
        """
        with SqlCursor(self.__db, commit) as sql:
            sql.execute("DELETE FROM tracks WHERE storage_type & ?",
                        (StorageType.COLLECTION,))

    def get_uris(self, uris_concerned=None):
        """
            Get all tracks uri
            @param uris_concerned as [uri as str]
            @return [str]
        """
        with SqlCursor(self.__db) as sql:
            uris = []
            if uris_concerned:
                for uri in uris_concerned:
                    result = sql.execute("SELECT uri\
                                          FROM tracks\
                                          WHERE uri LIKE ? AND\
                                          storage_type & ?",
                                         (uri + "%", StorageType.COLLECTION))
                    uris += list(itertools.chain(*result))
            else:
                result = sql.execute("SELECT uri FROM tracks\
                                      WHERE  storage_type & ?",
                                     (StorageType.COLLECTION,))
                uris = list(itertools.chain(*result))
            return uris

    def get_number(self, track_id):
        """
            Get track position in album
            @param track_id as int
            @return position as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT tracknumber FROM tracks\
                                  WHERE rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_lp_track_id(self, track_id):
        """
            Get Lollypop id
            @param track_id as int
            @return str
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT lp_track_id FROM\
                                  tracks where rowid=?",
                                 (track_id,))
            v = result.fetchone()
            if v and v[0]:
                return v[0]
            return ""

    def get_discnumber(self, track_id):
        """
            Get disc number for track id
            @param track_id as int
            @return discnumber as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT discnumber FROM tracks\
                                  WHERE rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_discname(self, track_id):
        """
            Get disc name for track id
            @param track_id as int
            @return discname as str
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT discname FROM tracks\
                                  WHERE rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return ""

    def get_duration(self, track_id):
        """
            Get track duration for track id
            @param track_id as int
            @return duration as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT duration FROM tracks\
                                  WHERE rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def set_duration(self, track_id, duration):
        """
            Get track duration for track id
            @param track_id as int
            @param duration as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks\
                         SET duration=?\
                         WHERE rowid=?", (duration, track_id,))

    def set_mtime(self, track_id, mtime):
        """
            Set track_mtime
            @param track_id as int
            @param mtime as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks SET mtime=? WHERE rowid=?",
                        (mtime, track_id))

    def is_empty(self):
        """
            Return True if no tracks in db
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT COUNT(1) FROM tracks  LIMIT 1")
            v = result.fetchone()
            if v is not None:
                return v[0] == 0
            return True

    def get_loved_track_ids(self, artist_ids, storage_type):
        """
            Get loved track ids
            @param storage_type as StorageType
            @return [int]
        """
        with SqlCursor(self.__db) as sql:
            filters = (LovedFlags.LOVED, storage_type)
            request = "SELECT tracks.rowid\
                       FROM tracks, album_artists, artists\
                       WHERE loved=? AND\
                       artists.rowid=album_artists.artist_id AND\
                       tracks.album_id=album_artists.album_id AND\
                       storage_type & ?"
            if artist_ids:
                filters += tuple(artist_ids)
                request += " AND "
                request += make_subrequest("album_artists.artist_id=?",
                                           "OR",
                                           len(artist_ids))
            request += " ORDER BY artists.name"
            result = sql.execute(request, filters)
            return list(itertools.chain(*result))

    def get_populars(self, artist_ids, storage_type, skipped, limit):
        """
            Return populars tracks
            @param artist_ids as int
            @param storage_type as StorageType
            @param skipped as bool
            @param limit as int
            @return track ids as [int]
        """
        with SqlCursor(self.__db) as sql:
            filters = (storage_type,)
            request = "SELECT tracks.rowid FROM"
            if artist_ids:
                request += " tracks, track_artists "
            else:
                request += " tracks "
            request += "WHERE rate >= 4 AND storage_type & ?"
            if artist_ids:
                filters += tuple(artist_ids)
                request += " AND track_artists.track_id=tracks.rowid AND"
                request += make_subrequest("track_artists.artist_id=?",
                                           "OR",
                                           len(artist_ids))
            if not skipped:
                request += " AND not loved &? "
                filters += (LovedFlags.SKIPPED,)
            filters += (limit,)
            request += " ORDER BY popularity DESC LIMIT ?"
            result = sql.execute(request, filters)
            track_ids = list(itertools.chain(*result))
            if len(track_ids) < limit:
                filters = (storage_type,)
                request = "SELECT tracks.rowid FROM"
                if artist_ids:
                    request += " tracks, track_artists "
                else:
                    request += " tracks "
                request += "WHERE popularity!=0 AND\
                            storage_type & ?"
                if artist_ids:
                    filters += tuple(artist_ids)
                    request += " AND track_artists.track_id=tracks.rowid AND"
                    request += make_subrequest("track_artists.artist_id=?",
                                               "OR",
                                               len(artist_ids))
                if not skipped:
                    request += " AND not loved &? "
                    filters += (LovedFlags.SKIPPED,)
                filters += (limit,)
                request += " ORDER BY popularity DESC LIMIT ?"
                result = sql.execute(request, filters)
                track_ids += list(itertools.chain(*result))
            return list(set(track_ids))

    def get_higher_popularity(self):
        """
            Get higher available popularity
            @return int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT popularity\
                                  FROM tracks\
                                  ORDER BY POPULARITY DESC LIMIT 1")
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_avg_popularity(self):
        """
            Return avarage popularity
            @return avarage popularity as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT AVG(popularity)\
                                  FROM (SELECT popularity\
                                        FROM tracks\
                                        ORDER BY POPULARITY DESC LIMIT 100)")
            v = result.fetchone()
            if v and v[0] is not None and v[0] > 5:
                return v[0]
            return 5

    def set_more_popular(self, track_id):
        """
            Increment popularity field
            @param track_id as int
            @raise sqlite3.OperationalError on db update
        """
        with SqlCursor(self.__db, True) as sql:
            result = sql.execute("SELECT popularity from tracks WHERE rowid=?",
                                 (track_id,))
            pop = result.fetchone()
            if pop:
                current = pop[0]
            else:
                current = 0
            current += 1
            sql.execute("UPDATE tracks set popularity=? WHERE rowid=?",
                        (current, track_id))

    def set_listened_at(self, track_id, time):
        """
            Set ltime for track
            @param track_id as int
            @param time as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks set ltime=? WHERE rowid=?",
                        (time, track_id))

    def get_little_played(self, storage_type, skipped, limit):
        """
            Return random tracks little played
            @param storage_type as StorageType
            @param skipped as bool
            @param limit as int
            @return tracks as [int]
        """
        with SqlCursor(self.__db) as sql:
            request = "SELECT rowid FROM tracks WHERE storage_type & ?"
            if not skipped:
                request += " AND loved !=-1 "
            request += " ORDER BY ltime, random() LIMIT ?"
            result = sql.execute(request, (storage_type, limit))
            return list(itertools.chain(*result))

    def get_recently_listened_to(self, storage_type, skipped, limit):
        """
            Return tracks listened recently
            @param storage_type as StorageType
            @param skipped as bool
            @param limit as int
            @return tracks as [int]
        """
        with SqlCursor(self.__db) as sql:
            filters = (storage_type,)
            request = "SELECT tracks.rowid FROM tracks\
                       WHERE ltime!=0 AND storage_type & ?"
            if not skipped:
                request += " AND not loved &? "
                filters += (LovedFlags.SKIPPED,)
            request += " ORDER BY ltime DESC LIMIT ?"
            filters += (limit,)
            result = sql.execute(request, filters)
            return list(itertools.chain(*result))

    def get_skipped(self, storage_type):
        """
            Return skipped tracks
            @param storage_type as StorageType
            @return tracks as [int]
        """
        with SqlCursor(self.__db) as sql:
            request = "SELECT rowid FROM tracks\
                       WHERE loved & ? AND storage_type & ?"
            result = sql.execute(request, (LovedFlags.SKIPPED, storage_type))
            return list(itertools.chain(*result))

    def get_randoms(self, genre_ids, storage_type, skipped, limit):
        """
            Return random tracks
            @param genre_ids as [int]
            @param storage_type as StorageType
            @parma skipped as bool
            @param limit as int
            @return track ids as [int]
        """
        with SqlCursor(self.__db) as sql:
            filters = (storage_type,)
            request = "SELECT tracks.rowid FROM tracks"
            if genre_ids:
                request += ",track_genres"
            request += " WHERE storage_type & ? "
            if not skipped:
                request += " AND not loved &? "
                filters += (LovedFlags.SKIPPED,)
            if genre_ids:
                request += "AND tracks.rowid=track_genres.track_id"
                filters += tuple(genre_ids)
                request += " AND "
                request += make_subrequest("track_genres.genre_id=?",
                                           "OR",
                                           len(genre_ids))
            request += " ORDER BY random() LIMIT ?"
            filters += (limit,)
            result = sql.execute(request, filters)
            return list(itertools.chain(*result))

    def set_popularity(self, track_id, popularity):
        """
            Set popularity
            @param track_id as int
            @param popularity as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks set popularity=? WHERE rowid=?",
                        (popularity, track_id))

    def get_popularity(self, track_id):
        """
            Get popularity
            @param track_id  as int
            @return popularity as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT popularity FROM tracks WHERE\
                                 rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_loved(self, track_id):
        """
            Get track loved status
            @param track_id as int
            @return loved as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT loved FROM tracks WHERE\
                                 rowid=?", (track_id,))

            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_year_for_album(self, album_id, disc_number=None):
        """
            Get album year based on tracks
            Use most used year by tracks
            @param album_id as int
            @param disc_number as int/None
            @return int
        """
        with SqlCursor(self.__db) as sql:
            filters = (album_id,)
            request = "SELECT year, COUNT(year) AS occurrence FROM tracks\
                       WHERE tracks.album_id=?"
            if disc_number is not None:
                filters += (disc_number,)
                request += " AND tracks.discnumber=?"
            request += " GROUP BY year\
                        ORDER BY occurrence DESC"
            result = sql.execute(request, filters)
            v = list(result)
            # Ignore album with multiple original date
            if len(v) == 1:
                return v[0][0]
            return None

    def get_ltime(self, track_id):
        """
            Get listen time
            @param track_id  as int
            @return listen time as int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT ltime FROM tracks WHERE\
                                 rowid=?", (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_mtime(self, track_id):
        """
            Get modification time
            @param track_id as int
            @return modification time as int
        """
        with SqlCursor(self.__db) as sql:
            request = "SELECT mtime FROM tracks\
                       WHERE tracks.rowid=?"
            result = sql.execute(request, (track_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def get_years(self, storage_type):
        """
            Return all tracks years and if unknown years exist
            @param storage_type as StorageType
            @return ([int], bool)
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT tracks.year\
                                  FROM tracks\
                                  WHERE storage_type & ?",
                                 (storage_type,))
            years = []
            unknown = False
            for year in list(itertools.chain(*result)):
                if year is None:
                    unknown = True
                elif year not in years:
                    years.append(year)
            return (years, unknown)

    def get_albums_by_disc_for_year(self, year, storage_type,
                                    skipped, limit=-1):
        """
            Return albums for year
            @param year as int
            @param storage_type as StorageType
            @param skipped as bool
            @param limit as int
            @return discs [(int, int)]
        """
        with SqlCursor(self.__db) as sql:
            order = " ORDER BY artists.sortname\
                     COLLATE NOCASE COLLATE LOCALIZED,\
                     tracks.timestamp,\
                     albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED LIMIT ?"
            request = "SELECT DISTINCT tracks.album_id,\
                       discnumber,\
                       discname,\
                       albums.year\
                       FROM albums, tracks, album_artists, artists\
                       WHERE albums.rowid=album_artists.album_id AND\
                       artists.rowid=album_artists.artist_id AND\
                       tracks.album_id=albums.rowid AND\
                       tracks.year=? AND albums.storage_type & ?"
            filters = (year, storage_type, limit)
            if not skipped:
                request += " AND not albums.loved &? "
                filters += (LovedFlags.SKIPPED,)
            request += " GROUP BY tracks.album_id"
            request += order
            result = sql.execute(request, filters)
            return list(result)

    def get_compilations_by_disc_for_year(self, year, storage_type,
                                          skipped, limit=-1):
        """
            Return compilations for year
            @param year as int
            @param storage_type as StorageType
            @param skipped as bool
            @param limit as int
            @return discs [(int, int)]
        """
        with SqlCursor(self.__db) as sql:
            order = " ORDER BY albums.timestamp, albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED LIMIT ?"
            request = "SELECT DISTINCT tracks.album_id,\
                       discnumber,\
                       discname,\
                       albums.year\
                       FROM albums, album_artists, tracks\
                       WHERE album_artists.artist_id=?\
                       AND album_artists.album_id=albums.rowid\
                       AND tracks.album_id=albums.rowid\
                       AND albums.storage_type & ?\
                       AND tracks.year=?"
            filters = (Type.COMPILATIONS, storage_type, year, limit)
            if not skipped:
                request += " AND not albums.loved &? "
                filters += (LovedFlags.SKIPPED,)
            request += " GROUP BY tracks.album_id"
            request += order
            result = sql.execute(request, filters)
            return list(result)

    def set_lp_track_id(self, track_id, lp_track_id):
        """
            Set lp track id
            @param album_id as int
            @param lp_album_id as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks SET lp_track_id=? WHERE rowid=?",
                        (lp_track_id, track_id))

    def set_loved(self, track_id, loved):
        """
            Set track loved
            @param track_id as int
            @param loved as int
            @warning: commit needed
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("UPDATE tracks SET loved=? WHERE rowid=?",
                        (loved, track_id))

    def count(self):
        """
            Count tracks
            @return int
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT COUNT(1) FROM tracks\
                                  WHERE storage_type & ?",
                                 (StorageType.COLLECTION | StorageType.SAVED,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return 0

    def clean(self, commit=True):
        """
            Clean database for track id
            @param commit as bool
        """
        with SqlCursor(self.__db, commit) as sql:
            sql.execute("DELETE FROM track_artists\
                         WHERE track_artists.track_id NOT IN (\
                            SELECT tracks.rowid FROM tracks)")
            sql.execute("DELETE FROM track_genres\
                         WHERE track_genres.track_id NOT IN (\
                            SELECT tracks.rowid FROM tracks)")

    def search(self, searched, storage_type):
        """
            Search for tracks looking like searched
            @param searched as str without accents
            @param storage_type as StorageType
            @return [(int, name)]
        """
        with SqlCursor(self.__db) as sql:
            filters = ("%" + searched + "%", storage_type)
            request = "SELECT rowid, name FROM tracks\
                       WHERE noaccents(name) LIKE ?\
                       AND tracks.storage_type & ? LIMIT 25"
            result = sql.execute(request, filters)
            return list(result)

    def search_performed(self, searched, storage_type):
        """
            Search tracks looking like searched with performers
            @param searched as str without accents
            @param storage_type as StorageType
            @return [(int, name)]
        """
        with SqlCursor(self.__db) as sql:
            filters = ("%" + searched + "%", storage_type)
            request = "SELECT DISTINCT tracks.rowid, artists.name\
                   FROM track_artists, tracks, artists\
                   WHERE track_artists.artist_id=artists.rowid AND\
                   track_artists.track_id=tracks.rowid AND\
                   noaccents(artists.name) LIKE ? AND\
                   tracks.storage_type & ? AND NOT EXISTS (\
                        SELECT album_artists.artist_id\
                        FROM album_artists\
                        WHERE album_artists.artist_id=artists.rowid)\
                    LIMIT 25"
            result = sql.execute(request, filters)
            return list(result)

    def search_track(self, artist, title):
        """
            Get track id for artist and title
            @param artist as string
            @param title as string
            @return track id as int
        """
        artist = noaccents(artist.lower())
        track_ids = self.get_ids_for_name(title)
        for track_id in track_ids:
            album_id = App().tracks.get_album_id(track_id)
            artist_ids = set(App().albums.get_artist_ids(album_id)) &\
                set(App().tracks.get_artist_ids(track_id))
            for artist_id in artist_ids:
                db_artist = noaccents(
                    App().artists.get_name(artist_id).lower())
                if artist.find(db_artist) != -1 or\
                        db_artist.find(artist) != -1:
                    return track_id
            artists = ", ".join(App().tracks.get_artists(track_id)).lower()
            if noaccents(artists) == artist:
                return track_id
        return None

    def remove(self, track_id):
        """
            Remove track
            @param track_id as int
        """
        with SqlCursor(self.__db, True) as sql:
            sql.execute("DELETE FROM track_genres\
                         WHERE track_id=?", (track_id,))
            sql.execute("DELETE FROM track_artists\
                         WHERE track_id=?", (track_id,))
            sql.execute("DELETE FROM tracks\
                         WHERE rowid=?", (track_id,))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from gettext import gettext as _
import json
from locale import getdefaultlocale

from lollypop.define import App
from lollypop.utils import escape
from lollypop.logger import Logger


class WikipediaHelper:
    """
        Helper for wikipedia search
    """

    __API_SEARCH = "https://%s.wikipedia.org/w/api.php?action=query" +\
        "&list=search&srsearch=%s&format=json"
    __API_INFO = "https://%s.wikipedia.org/w/api.php?action=query" +\
        "&pageids=%s&format=json" +\
        "&prop=extracts&exlimit=max&explaintext&redirects=1"

    def __init__(self):
        """
            Init wikipedia
        """
        self.__locale = getdefaultlocale()[0][0:2]

    def get_content_for_term(self, term):
        """
            Get content for term
            @param term as str
            @return bytes/None
        """
        try:
            (locale, page_id) = self.__search_term(term)
            if page_id is None:
                return None
            uri = self.__API_INFO % (locale, page_id)
            (status, data) = App().task_helper.load_uri_content_sync(uri)
            if status:
                decode = json.loads(data.decode("utf-8"))
                extract = decode["query"]["pages"][str(page_id)]["extract"]
                return extract.encode("utf-8")
        except Exception as e:
            Logger.error("Wikipedia::get_content_for_term(): %s", e)
        return None

    def get_content_for_page_id(self, page_id, locale):
        """
            Get page content
            @param page_id as str
            @param locale as str
            @return bytes/None
        """
        try:
            uri = self.__API_INFO % (locale, page_id)
            (status, data) = App().task_helper.load_uri_content_sync(uri)
            if status:
                decode = json.loads(data.decode("utf-8"))
                extract = decode["query"]["pages"][str(page_id)]["extract"]
                return extract.encode("utf-8")
        except Exception as e:
            Logger.error("Wikipedia::get_content_for_page_id(): %s", e)
        return None

    def get_search_list(self, term):
        """
            Get search list for term
            @param term as str
            @return [(str, str)]: list of locales/title
        """
        pages = []
        try:
            for locale in [self.__locale, "en"]:
                uri = self.__API_SEARCH % (locale, term)
                (status, data) = App().task_helper.load_uri_content_sync(uri)
                decode = json.loads(data.decode("utf-8"))
                if status:
                    for item in decode["query"]["search"]:
                        pages.append((locale, item["title"], item["pageid"]))
        except Exception as e:
            print("Wikipedia::get_search_list(): %s", e)
        return pages

#######################
# PRIVATE             #
#######################
    def __search_term(self, term):
        """
            Search term on Wikipdia
            @param term as str
            @return pageid as str
        """
        try:
            for locale in [self.__locale, "en"]:
                uri = self.__API_SEARCH % (locale, term)
                (status, data) = App().task_helper.load_uri_content_sync(uri)
                if status:
                    decode = json.loads(data.decode("utf-8"))
                    for item in decode["query"]["search"]:
                        if escape(item["title"].lower()) ==\
                                escape(term.lower()):
                            return (locale, item["pageid"])
                        else:
                            for word in [_("band"), _("singer"),
                                         "band", "singer"]:
                                if item["snippet"].lower().find(word) != -1:
                                    return (locale, item["pageid"])
        except Exception as e:
            print("Wikipedia::__search_term(): %s", e)
        return ("", None)



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
from time import time
from random import shuffle
from locale import getdefaultlocale

from lollypop.logger import Logger
from lollypop.utils import get_default_storage_type
from lollypop.helper_web_spotify import SpotifyWebHelper
from lollypop.define import App, StorageType


class SpotifyCollectionWebService(SpotifyWebHelper):
    """
        Add items to collection with Spotify
        Depends on SaveWebHelper
    """
    def __init__(self):
        """
            Init object
        """
        SpotifyWebHelper.__init__(self)

    def search_similar_albums(self, cancellable):
        """
            Add similar albums to DB
            @param cancellable as Gio.Cancellable
        """
        Logger.info("Get similar albums from Spotify")
        from lollypop.similars_spotify import SpotifySimilars
        similars = SpotifySimilars()
        try:
            storage_type = get_default_storage_type()
            artists = App().artists.get_randoms(
                self.MAX_ITEMS_PER_STORAGE_TYPE, storage_type)
            artist_names = [name for (aid, name, sortname) in artists]
            similar_ids = similars.get_similar_artist_ids(artist_names,
                                                          cancellable)
            # Add albums
            shuffle(similar_ids)
            for similar_id in similar_ids[:self.MAX_ITEMS_PER_STORAGE_TYPE]:
                albums_payload = self.__get_artist_albums_payload(similar_id,
                                                                  cancellable)
                shuffle(albums_payload)
                for album in albums_payload:
                    if cancellable.is_cancelled():
                        raise Exception("Cancelled")
                    lollypop_payload = SpotifyWebHelper.lollypop_album_payload(
                        self, album)
                    self.save_album_payload_to_db(
                                           lollypop_payload,
                                           StorageType.SPOTIFY_SIMILARS,
                                           True,
                                           cancellable)
                    break
        except Exception as e:
            Logger.warning("SpotifyWebService::search_similar_albums(): %s", e)

    def search_new_releases(self, cancellable):
        """
            Get new released albums from spotify
            @param cancellable as Gio.Cancellable
        """
        Logger.info("Get new releases from Spotify")
        try:
            locale = getdefaultlocale()[0][0:2]
            token = App().ws_director.token_ws.get_token("SPOTIFY",
                                                         cancellable)
            bearer = "Bearer %s" % token
            headers = [("Authorization", bearer)]
            uri = "https://api.spotify.com/v1/browse/new-releases"
            uris = ["%s?country=%s" % (uri, locale), uri]
            for uri in uris:
                if cancellable.is_cancelled():
                    raise Exception("cancelled")
                (status,
                 data) = App().task_helper.load_uri_content_sync_with_headers(
                    uri, headers, cancellable)
                if status:
                    decode = json.loads(data.decode("utf-8"))
                    for album in decode["albums"]["items"]:
                        if cancellable.is_cancelled():
                            raise Exception("Cancelled")
                        lollypop_payload =\
                            SpotifyWebHelper.lollypop_album_payload(
                                self, album)
                        self.save_album_payload_to_db(
                                             lollypop_payload,
                                             StorageType.SPOTIFY_NEW_RELEASES,
                                             True,
                                             cancellable)
                    # Check if storage type needs to be updated
                    # Check if albums newer than a week are enough
                    timestamp = time() - 604800
                    newer_albums = App().albums.get_newer_for_storage_type(
                                             StorageType.SPOTIFY_NEW_RELEASES,
                                             timestamp)
                    if len(newer_albums) >= self.MIN_ITEMS_PER_STORAGE_TYPE:
                        break
        except Exception as e:
            Logger.warning("SpotifyWebService::search_new_releases(): %s", e)

#######################
# PRIVATE             #
#######################
    def __get_artist_albums_payload(self, spotify_id, cancellable):
        """
            Get albums payload for artist
            @param spotify_id as str
            @param cancellable as Gio.Cancellable
            @return {}
        """
        try:
            token = App().ws_director.token_ws.get_token("SPOTIFY",
                                                         cancellable)
            bearer = "Bearer %s" % token
            headers = [("Authorization", bearer)]
            uri = "https://api.spotify.com/v1/artists/%s/albums" % spotify_id
            (status,
             data) = App().task_helper.load_uri_content_sync_with_headers(
                    uri, headers, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                return decode["items"]
        except Exception as e:
            Logger.warning(
                "SpotifyWebService::__get_artist_albums_payload(): %s", e)
        return None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, GLib

from gettext import gettext as _

from lollypop.utils import get_default_storage_type
from lollypop.define import App, MARGIN, ViewType
from lollypop.helper_horizontal_scrolling import HorizontalScrollingHelper
from lollypop.view_artists_rounded import RoundedArtistsView


class ArtistsLineView(RoundedArtistsView, HorizontalScrollingHelper):
    """
        Show artist in an horizontal flowbox
    """

    def __init__(self, storage_type, view_type):
        """
            Init artist view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        RoundedArtistsView.__init__(self, storage_type, view_type)
        self.__artist_ids = []
        self.set_row_spacing(5)
        self._label = Gtk.Label.new()
        self._label.set_ellipsize(Pango.EllipsizeMode.END)
        self._label.get_style_context().add_class("dim-label")
        self.__update_label(App().window.folded)
        self._label.set_hexpand(True)
        self._label.set_property("halign", Gtk.Align.START)
        self._backward_button = Gtk.Button.new_from_icon_name(
                                                    "go-previous-symbolic",
                                                    Gtk.IconSize.BUTTON)
        self._forward_button = Gtk.Button.new_from_icon_name(
                                                   "go-next-symbolic",
                                                   Gtk.IconSize.BUTTON)
        self._backward_button.get_style_context().add_class("menu-button")
        self._forward_button.get_style_context().add_class("menu-button")
        header = Gtk.Grid()
        header.set_column_spacing(10)
        header.add(self._label)
        header.add(self._backward_button)
        header.add(self._forward_button)
        header.set_margin_end(MARGIN)
        header.show_all()
        HorizontalScrollingHelper.__init__(self)
        self.add(header)
        self._label.set_property("halign", Gtk.Align.START)
        self.add_widget(self._box)

    def add_value(self, artist_id):
        """
            Insert item if not exists
            @param artist_id as int
        """
        if artist_id not in self.__artist_ids:
            self.__artist_ids.append(artist_id)
            RoundedArtistsView.add_value(self, artist_id)
            self.update_buttons()

    def clear(self):
        """
            Clear children
        """
        self.__artist_ids = []
        RoundedArtistsView.clear(self)

    @property
    def args(self):
        return None

#######################
# PROTECTED           #
#######################
    def _on_container_folded(self, leaflet, folded):
        """
            Handle libhandy folded status
            @param leaflet as Handy.Leaflet
            @param folded as Gparam
        """
        RoundedArtistsView._on_container_folded(self, leaflet, folded)
        self.__update_label(folded)

    def _on_collection_updated(self, scanner, item, scan_update):
        pass

    def _on_populated(self, widget):
        """
            Update buttons
            @param widget as Gtk.Widget
        """
        self.update_buttons()
        RoundedArtistsView._on_populated(self, widget)

#######################
# PRIVATE             #
#######################
    def __update_label(self, folded):
        """
            Update label style based on current adaptive state
            @param folded as bool
        """
        style_context = self._label.get_style_context()
        if folded:
            style_context.remove_class("text-x-large")
        else:
            style_context.add_class("text-x-large")


class ArtistsRandomLineView(ArtistsLineView):
    """
        Line view showing 6 random artists
    """
    def __init__(self, storage_type, view_type):
        """
            Init artist view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        ArtistsLineView.__init__(self, storage_type, view_type)
        self._label.set_text(_("Why not listen to?"))

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            self._box.set_min_children_per_line(len(items))
            ArtistsLineView.populate(self, items)
            if items:
                self.show()

        def load():
            storage_type = get_default_storage_type()
            ids = App().artists.get_randoms(15, storage_type)
            return ids

        App().task_helper.run(load, callback=(on_load,))


class ArtistsSearchLineView(ArtistsLineView):
    """
        Line view for search
    """
    def __init__(self, storage_type):
        """
            Init artist view
            @param storage_type as StorageType
        """
        ArtistsLineView.__init__(self, storage_type,
                                 ViewType.SEARCH | ViewType.SCROLLED)
        self.__artist_ids = []
        self._label.set_text(_("Artists"))

    def add_value(self, item_id):
        """
            Insert item
            @param item_id as int
        """
        if item_id in self.__artist_ids:
            return
        self.__artist_ids.append(item_id)
        ArtistsLineView.populate(self, [item_id])
        self._box.set_min_children_per_line(len(self._box.get_children()))

    def clear(self):
        """
            Clear and hide the view
        """
        self.__artist_ids = []
        ArtistsLineView.clear(self)
        GLib.idle_add(self.hide)



# Copyright (c) 2014-2016 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import itertools

from lollypop.sqlcursor import SqlCursor
from lollypop.define import Lp, Type


class MpdDatabase:
    """
        Databse request from MPD module
    """

    def count(self, album, artist_id, genre_id, year):
        """
            Count songs and play time
            @param album as string
            @param artist_id as int
            @param genre_id as int
            @param year as int
        """
        songs = 0
        playtime = 0
        with SqlCursor(Lp().db) as sql:
            result = self._get_tracks(sql, "COUNT(*), SUM(tracks.duration)",
                                      album, artist_id, genre_id, year)
            v = result.fetchone()
            if v is not None:
                if v[0] is not None:
                    songs = v[0]
                if v[1] is not None:
                    playtime = v[1]
            return (songs, playtime)

    def get_tracks_paths(self, album, artist_id, genre_id, year):
        """
            Get tracks path
            @param album as string
            @param artist_id as int
            @param genre_id as int
            @param year as int
            @return paths as [str]
        """
        with SqlCursor(Lp().db) as sql:
            result = self._get_tracks(sql, "filepath",
                                      album, artist_id, genre_id, year)
            return list(itertools.chain(*result))

    def get_tracks_ids(self, album, artist_id, genre_id, year):
        """
            Get tracks path
            @param album as string
            @param artist_id as int
            @param genre_id as int
            @param year as int
            @return paths as [str]
        """
        with SqlCursor(Lp().db) as sql:
            result = self._get_tracks(sql, "tracks.rowid",
                                      album, artist_id, genre_id, year)
            return list(itertools.chain(*result))

    def get_albums_names(self, artist_id, genre_id, year):
        """
            Get albums names
            @param artist_id as int
            @param genre_id as int
            @param year as int
            @return names as [str]
        """
        from_str = "albums "
        where_str = ""
        if artist_id is not None:
            where_str += " albums.artist_id = %s AND" % artist_id
        if genre_id is not None:
            from_str += ",album_genres"
            where_str += " album_genres.genre_id = %s\
                          AND album_genres.album_id = albums.rowid\
                          AND" % genre_id
        if year is None:
            where_str += " albums.year is null"
        elif year != Type.NONE:
            where_str += " albums.year = %s" % year

        with SqlCursor(Lp().db) as sql:
            request = "SELECT albums.name FROM "\
                       + from_str
            if where_str != "":
                request += " WHERE " + where_str
            if request.endswith("AND"):
                request = request[:-3]
            result = sql.execute(request)
            return list(itertools.chain(*result))

    def get_artists_names(self, genre_id):
        """
            Get artists names
            @param genre_id as int
            @return names as [str]
        """
        from_str = "artists, albums "
        where_str = " albums.artist_id = artists.rowid AND"
        if genre_id is not None:
            from_str += ", album_genres"
            where_str += " album_genres.genre_id = %s\
                          AND album_genres.album_id = albums.rowid\
                          AND" % genre_id

        with SqlCursor(Lp().db) as sql:
            request = "SELECT DISTINCT artists.name FROM "\
                       + from_str
            if where_str != "":
                request += " WHERE " + where_str
            if request.endswith("AND"):
                request = request[:-3]
            result = sql.execute(request)
            return list(itertools.chain(*result))

    def get_albums_years(self, album, artist_id, genre_id):
        """
            Get all availables albums years
            @param album as string
            @param artist_id as int
            @param genre_id as int
            @return years as [str]
        """
        from_str = "albums "
        where_str = "year is not null AND"
        if album is not None:
            where_str += ' albums.name = "%s" AND' % album
        if artist_id is not None:
            where_str += " albums.artist_id= %s AND" % artist_id
        if genre_id is not None:
            from_str += ",album_genres "
            where_str += " album_genres.genre_id = %s\
                          AND album_genres.album_id = albums.rowid\
                          AND" % genre_id
        request = "SELECT year FROM " + from_str
        if where_str != "":
            request += " WHERE " + where_str
        if request.endswith("AND"):
            request = request[:-3]
        with SqlCursor(Lp().db) as sql:
            result = sql.execute(request)
            return list(itertools.chain(*result))

    def listallinfos(self):
        """
            Get all tracks
            @return array [(
                     track.path,
                     track.artist,
                     track.album.name,
                     track.album_artist,
                     track.name,
                     track.album.year,
                     track.genre,
                     track.duration,
                     track.id,
                     track.position)]
        """
        request = "SELECT tracks.filepath, artists.name, albums.name,\
                   artists.name, tracks.name, albums.year, genres.name,\
                   tracks.duration, tracks.rowid, tracks.tracknumber\
                   FROM artists, albums, tracks, genres, track_genres\
                   WHERE albums.rowid = tracks.album_id\
                   AND artists.rowid = albums.artist_id\
                   AND genres.rowid = track_genres.genre_id\
                   AND tracks.rowid = track_genres.track_id"
        with SqlCursor(Lp().db) as sql:
            result = sql.execute(request)
            return list(result)

#######################
# PRIVATE             #
#######################
    def _get_tracks(self, sql, select_str, album, artist_id, genre_id, year):
        """
            Get tracks attributes
            @param sql as sqlite cursor
            @param select as string
            @param album as string
            @param artist_id as int
            @param genre_id as int
            @param year as int
            @return sqlite cursor
        """
        from_str = "tracks "
        where_str = ""
        if album is not None:
            from_str += ",albums"
            where_str += 'albums.name = "%s" AND\
                          tracks.album_id = albums.rowid AND' % album
        if artist_id is not None:
            from_str += ", artists"
            if "albums" not in from_str:
                from_str += ",albums"
                where_str += " tracks.album_id = albums.rowid AND"
            where_str += " artists.rowid = %s\
                          AND albums.artist_id = artists.rowid\
                          AND" % artist_id
        if genre_id is not None:
            from_str += ",track_genres"
            where_str += " track_genres.genre_id = %s\
                          AND track_genres.track_id = tracks.rowid\
                          AND" % genre_id
        if year is None:
            where_str += " tracks.year is null"
        elif year != Type.NONE:
            where_str += " tracks.year = %s" % year

        request = "SELECT %s FROM " % select_str\
                  + from_str
        if where_str != "":
            request += " WHERE " + where_str
        if request.endswith("AND"):
            request = request[:-3]
        result = sql.execute(request + "ORDER BY tracks.tracknumber")
        return result



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GdkPixbuf, GLib, GObject

from PIL import Image, ImageFilter

from lollypop.define import CACHE_PATH
from lollypop.define import App, StoreExtention, ArtSize, ArtBehaviour
from lollypop.utils_file import create_dir


class ArtworkManager(GObject.GObject):
    """
        Common methods for artworks manager
    """

    __gsignals__ = {
        "artwork-cleared": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        "album-artwork-changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "artist-artwork-changed": (GObject.SignalFlags.RUN_FIRST,
                                   None, (str,)),
        "uri-artwork-found": (GObject.SignalFlags.RUN_FIRST, None,
                              (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self):
        """
            Init artwork manager
        """
        GObject.GObject.__init__(self)
        create_dir(CACHE_PATH)
        App().settings.connect("changed::hd-artwork",
                               self.__on_hd_artwork_changed)
        self.__on_hd_artwork_changed()

    def save_pixbuf(self, pixbuf, path):
        """
            Save pixbuf at path
            @param pixbuf as GdkPixbuf.Pixbuf
            @param path as str
        """
        if self.__extension == StoreExtention.PNG:
            pixbuf.savev(path, "png", [None], [None])
        else:
            pixbuf.savev(path, "jpeg", ["quality"], ["100"])

    def add_extension(self, path):
        """
            Add file extension to path
            @param path as str
            @return str
        """
        if self.__extension == StoreExtention.PNG:
            return "%s.png" % path
        else:
            return "%s.jpg" % path

    def load_behaviour(self, pixbuf, width, height, behaviour):
        """
            Load behaviour on pixbuf
            @param width as int
            @param height as int
            @param behaviour as ArtBehaviour
        """
        # Crop image as square
        if behaviour & ArtBehaviour.CROP_SQUARE:
            pixbuf = self._crop_pixbuf_square(pixbuf)
        # Crop image keeping ratio
        elif behaviour & ArtBehaviour.CROP:
            pixbuf = self._crop_pixbuf(pixbuf, width, height)

        # Handle blur
        if behaviour & ArtBehaviour.BLUR:
            pixbuf = pixbuf.scale_simple(width,
                                         height,
                                         GdkPixbuf.InterpType.NEAREST)
            pixbuf = self._get_blur(pixbuf, 25)
        elif behaviour & ArtBehaviour.BLUR_HARD:
            pixbuf = pixbuf.scale_simple(width,
                                         height,
                                         GdkPixbuf.InterpType.NEAREST)
            pixbuf = self._get_blur(pixbuf, 50)
        elif behaviour & ArtBehaviour.BLUR_MAX:
            pixbuf = pixbuf.scale_simple(width,
                                         height,
                                         GdkPixbuf.InterpType.NEAREST)
            pixbuf = self._get_blur(pixbuf, 100)
        else:
            pixbuf = pixbuf.scale_simple(width,
                                         height,
                                         GdkPixbuf.InterpType.BILINEAR)
        return pixbuf

    def update_art_size(self):
        """
            Update value with some check
        """
        value = App().settings.get_value("cover-size").get_int32()
        # Check value as user can enter bad value via dconf
        if value < 50 or value > 400:
            value = 200
        ArtSize.BIG = value
        ArtSize.BANNER = int(ArtSize.BIG * 150 / 200)
        ArtSize.MEDIUM = int(ArtSize.BIG * 100 / 200)
        ArtSize.SMALL = int(ArtSize.BIG * 50 / 200)

    def save_pixbuf_from_data(self, store_path, data,
                              width=-1, height=-1):
        """
            Save a pixbuf at path from data
            @param store path as str
            @param data as bytes
            @param width as int
            @param height as int
        """
        if data is None:
            f = Gio.File.new_for_path(store_path)
            fstream = f.replace(None, False,
                                Gio.FileCreateFlags.REPLACE_DESTINATION, None)
            fstream.close()
        else:
            bytes = GLib.Bytes.new(data)
            stream = Gio.MemoryInputStream.new_from_bytes(bytes)
            pixbuf = GdkPixbuf.Pixbuf.new_from_stream(stream, None)
            if width != -1 and height != -1:
                pixbuf = pixbuf.scale_simple(width,
                                             height,
                                             GdkPixbuf.InterpType.BILINEAR)
            stream.close()
            self.save_pixbuf(pixbuf, store_path)

    @property
    def extension(self):
        """
            Get current artwork extension
            @return StoreExtention
        """
        return self.__extension

    @property
    def extension_str(self):
        """
            Get current artwork extension
            @return str
        """
        if self.__extension == StoreExtention.PNG:
            return "png"
        else:
            return "jpg"

#######################
# PROTECTED           #
#######################
    def _crop_pixbuf(self, pixbuf, wanted_width, wanted_height):
        """
            Crop pixbuf
            @param pixbuf as GdkPixbuf.Pixbuf
            @param wanted_width as int
            @param wanted height as int
            @return GdkPixbuf.Pixbuf
        """
        width = pixbuf.get_width()
        height = pixbuf.get_height()
        aspect = width / height
        wanted_aspect = wanted_width / wanted_height
        if aspect > wanted_aspect:
            new_width = height * wanted_aspect
            offset = (width - new_width)
            new_pixbuf = pixbuf.new_subpixbuf(offset / 2,
                                              0,
                                              width - offset,
                                              height)
        else:
            new_height = width / wanted_aspect
            offset = (height - new_height)
            new_pixbuf = pixbuf.new_subpixbuf(0,
                                              offset / 2,
                                              width,
                                              height - offset)
        return new_pixbuf

    def _crop_pixbuf_square(self, pixbuf):
        """
            Crop pixbuf as square
            @param pixbuf as GdkPixbuf.Pixbuf
            @return GdkPixbuf.Pixbuf
        """
        width = pixbuf.get_width()
        height = pixbuf.get_height()
        if width == height:
            new_pixbuf = pixbuf
        elif width > height:
            diff = (width - height)
            new_pixbuf = pixbuf.new_subpixbuf(diff / 2,
                                              0,
                                              width - diff,
                                              height)
        else:
            diff = (height - width)
            new_pixbuf = pixbuf.new_subpixbuf(0,
                                              diff / 2,
                                              width,
                                              height - diff)
        return new_pixbuf

    def _get_blur(self, pixbuf, gaussian):
        """
            Blur surface using PIL
            @param pixbuf as GdkPixbuf.Pixbuf
            @param gaussian as int
            @return GdkPixbuf.Pixbuf
        """
        if pixbuf is None:
            return None
        width = pixbuf.get_width()
        height = pixbuf.get_height()
        data = pixbuf.get_pixels()
        stride = pixbuf.get_rowstride()
        has_alpha = pixbuf.get_has_alpha()
        if has_alpha:
            mode = "RGBA"
            dst_row_stride = width * 4
        else:
            mode = "RGB"
            dst_row_stride = width * 3
        tmp = Image.frombytes(mode, (width, height),
                              data, "raw", mode, stride)
        tmp = tmp.filter(ImageFilter.GaussianBlur(gaussian))
        bytes = GLib.Bytes.new(tmp.tobytes())
        del pixbuf
        pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(bytes,
                                                 GdkPixbuf.Colorspace.RGB,
                                                 has_alpha,
                                                 8,
                                                 width,
                                                 height,
                                                 dst_row_stride)
        return pixbuf

#######################
# PRIVATE             #
#######################
    def __on_hd_artwork_changed(self, *ignore):
        """
            Update extension value
        """
        if App().settings.get_value("hd-artwork"):
            self.__extension = StoreExtention.PNG
        else:
            self.__extension = StoreExtention.JPG



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk, GObject

from lollypop.define import ArtSize, ViewType, MARGIN, App
from lollypop.utils import emit_signal
from lollypop.helper_size_allocation import SizeAllocationHelper


class Overlay(Gtk.Overlay):
    """
        Overlay with constraint size
    """
    def __init__(self):
        """
            Init overlay
            @param banner as BannerWidget
        """
        Gtk.Overlay.__init__(self)

    def do_get_preferred_width(self):
        """
            Force preferred width
        """
        (min, nat) = Gtk.Bin.do_get_preferred_width(self)
        # Allow resizing
        return (1, 1)


class BannerWidget(Gtk.Revealer, SizeAllocationHelper):
    """
        Default banner widget
    """

    gsignals = {
        "scroll": (GObject.SignalFlags.RUN_FIRST, None, (float, float)),
        "height-changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }
    for signal in gsignals:
        args = gsignals[signal]
        GObject.signal_new(signal, Gtk.Revealer,
                           args[0], args[1], args[2])

    def __init__(self, view_type):
        """
            Init bannner
            @param view_type as ViewType
        """
        Gtk.Revealer.__init__(self)
        self._artwork = None
        self.__view_type = view_type | ViewType.BANNER
        self.__width = 1
        self.set_property("valign", Gtk.Align.START)
        if view_type & ViewType.OVERLAY:
            self._overlay = Overlay()
            self._overlay.show()
            self._artwork = Gtk.Image()
            self._artwork.show()
            if App().animations:
                self._artwork.set_opacity(0.99)
            self.get_style_context().add_class("default-banner")
            self._artwork.get_style_context().add_class("default-banner")
            eventbox = Gtk.EventBox.new()
            eventbox.show()
            eventbox.add_events(Gdk.EventMask.ALL_EVENTS_MASK)
            eventbox.add(self._artwork)
            self._overlay.add(eventbox)
            self.__event_controller = Gtk.EventControllerScroll.new(
                eventbox, Gtk.EventControllerScrollFlags.BOTH_AXES)
            self.__event_controller.set_propagation_phase(
                Gtk.PropagationPhase.TARGET)
            self.__event_controller.connect("scroll", self.__on_scroll)
            self.add(self._overlay)
        self.set_reveal_child(True)
        self.set_transition_duration(250)
        self.connect("destroy", self.__on_destroy)
        SizeAllocationHelper.__init__(self)

    def update_for_width(self, width):
        """
            Update banner internals for width, call this before showing banner
            @param width as int
        """
        self.__width = width
        if self._artwork is not None:
            self._artwork.set_size_request(-1, self.height)

    @property
    def width(self):
        """
            Get current width
            @return int
        """
        # If not SMALL, we add sidebar width because we want banners
        # to calculate sizing with sidebar included. Allows to prevent
        # glitch when sidebar is auto shown
        if not App().window.folded:
            sidebar_width = App().window.container.sidebar.\
                get_allocated_width()
            return self.__width + sidebar_width
        else:
            return self.__width

    @property
    def height(self):
        """
            Get wanted height
            @return int
        """
        if App().window.folded:
            height = ArtSize.MEDIUM + MARGIN * 2
        else:
            height = ArtSize.BANNER + MARGIN * 2
        return height

    @property
    def view_type(self):
        """
            Get view type
            @return ViewType
        """
        return self.__view_type

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Update artwork
            @param allocation as Gtk.Allocation
            @return bool
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            if allocation.width != self.__width:
                self.__width = allocation.width
                height = self.height
                if self._artwork is not None:
                    self._artwork.set_size_request(-1, height)
                emit_signal(self, "height-changed", height)
                return True
        return False

    def _on_artwork(self, surface):
        """
            Set album artwork
            @param surface as str
        """
        banner_context = self.get_style_context()
        artwork_context = self._artwork.get_style_context()
        self._artwork.set_from_surface(surface)
        if surface is None:
            banner_context.add_class("default-banner")
            banner_context.remove_class("black")
            artwork_context.add_class("default-banner")
        else:
            self.emit("populated")
            banner_context.remove_class("default-banner")
            banner_context.add_class("black")
            artwork_context.remove_class("default-banner")

#######################
# PRIVATE             #
#######################
    def __on_destroy(self, widget):
        """
            Remove ref cycle
            @param widget as Gtk.Widget
        """
        self.__event_controller = None

    def __on_scroll(self, event_controller, x, y):
        """
            Pass scroll
            @param event_controller as Gtk.EventControllerScroll
            @param x as int
            @param y as int
        """
        emit_signal(self, "scroll", x, y)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

import sqlite3
from threading import Lock

from lollypop.sqlcursor import SqlCursor


class History:
    """
        History manager, allow Lollypop to restore stats when a file is moved
    """
    __LOCAL_PATH = GLib.get_user_data_dir() + "/lollypop"
    __DB_PATH = "%s/history.db" % __LOCAL_PATH
    __LIMIT = 1000000  # Delete when limit is reached
    __DELETE = 100     # How many elements to delete
    __create_history = """CREATE TABLE history (
                            id INTEGER PRIMARY KEY,
                            name TEXT NOT NULL,
                            duration INT NOT NULL,
                            ltime INT NOT NULL,
                            popularity INT NOT NULL,
                            rate INT NOT NULL,
                            mtime INT NOT NULL,
                            album_rate INT NOT NULL,
                            loved INT NOT NULL,
                            album_loved INT NOT NULL,
                            album_synced INT NOT NULL,
                            album_popularity INT NOT NULL)"""

    def __init__(self):
        """
            Init playlists manager
        """
        self.thread_lock = Lock()
        # Create db schema
        try:
            with SqlCursor(self, True) as sql:
                sql.execute(self.__create_history)
        except:
            pass
        with SqlCursor(self, True) as sql:
            result = sql.execute("SELECT COUNT(*)\
                                  FROM history")
            v = result.fetchone()
            if v is not None and v[0] > self.__LIMIT:
                sql.execute("DELETE FROM history\
                             WHERE rowid IN (SELECT rowid\
                                             FROM history\
                                             LIMIT %s)" % self.__DELETE)
                sql.execute("VACUUM")

    def add(self, name, duration, popularity, rate, ltime, mtime, loved,
            album_loved, album_popularity, album_rate, album_synced):
        """
            Add an item to history
            @param name as str
            @param duration as int
            @param popularity as int
            @param rate as int
            @param ltime as int
            @param mtime as int
            @param loved as bool
            @param album_loved as bool
            @param album_popularity as int
            @param album_rate as int
            @param album_synced as int
            @thread safe
        """
        with SqlCursor(self, True) as sql:
            # Needed because of seconds to ms DB migration
            # Value in DB is rounded version of Gstreamer value
            duration //= 1000
            if self.exists(name, duration):
                sql.execute("UPDATE history\
                             SET popularity=?,rate=?,ltime=?,mtime=?,loved=?,\
                             album_loved=?,album_popularity=?,album_rate=?,\
                             album_synced=?\
                             WHERE name=? AND duration=?",
                            (popularity, rate, ltime, mtime, loved,
                             album_loved, album_popularity, album_rate,
                             album_synced,
                             name, duration))
            else:
                sql.execute("INSERT INTO history\
                             (name, duration, popularity, rate, ltime, mtime,\
                             loved, album_loved, album_popularity, album_rate,\
                             album_synced)\
                             VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                            (name, duration, popularity, rate, ltime, mtime,
                             loved, album_loved, album_popularity, album_rate,
                             album_synced))

    def get(self, name, duration):
        """
            Get stats for track with name and duration
            @param name as str
            @param duration as int
            @return (popularity, ltime, mtime,
                     loved album, album_popularity, album_synced)
             as (int, int, int, int, int, int)
        """
        with SqlCursor(self) as sql:
            duration //= 1000
            result = sql.execute("SELECT popularity, rate, ltime, mtime,\
                                  loved, album_loved, album_popularity,\
                                  album_rate, album_synced\
                                  FROM history\
                                  WHERE name=?\
                                  AND duration=?",
                                 (name, duration))
            v = result.fetchone()
            if v is not None:
                return v
            return (0, 0, 0, 0, 0, 0, 0, 0, 0)

    def exists(self, name, duration):
        """
            Return True if entry exists
            @param name as str
            @param duration as int
            @return bool
        """
        with SqlCursor(self) as sql:
            result = sql.execute("SELECT rowid\
                                  FROM history\
                                  WHERE name=?\
                                  AND duration=?",
                                 (name, duration))
            v = result.fetchone()
            if v is not None:
                return True
            else:
                return False

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            return sqlite3.connect(self.__DB_PATH, 600.0)
        except:
            exit(-1)

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from gettext import gettext as _

from lollypop.define import ViewType


class DecadeMenu(Gio.Menu):
    """
        Contextual menu for a decade
    """
    def __init__(self, years, view_type, header=False):
        """
            Init decade menu
            @param years as [int]
            @param view_type as ViewType
            @param header as bool
        """
        Gio.Menu.__init__(self)
        if header:
            from lollypop.menu_header import RoundedMenuHeader
            name = "%s - %s" % (years[0], years[-1])
            artwork_name = "decade_%s" % name
            self.append_item(RoundedMenuHeader(name, artwork_name))
        if not view_type & ViewType.BANNER:
            from lollypop.menu_playback import DecadePlaybackMenu
            self.append_section(_("Playback"), DecadePlaybackMenu(years))



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json

from lollypop.logger import Logger
from lollypop.define import App


class MusicBrainzWebHelper:
    """
        Web helper for MusicBrainz
    """

    def __init__(self):
        """
            Init helper
        """
        pass

    def get_tracks_payload(self, mbid, cancellable):
        """
            Get tracks payload for album mbid
            @param mbid as str
            @param cancellable as Gio.Cancellable
            @return payload as []
        """
        try:
            uri = "http://musicbrainz.org/ws/2/release/%s" % mbid
            uri += "?inc=recordings+artist-credits&fmt=json"
            (status, data) = App().task_helper.load_uri_content_sync(uri, None)
            if status:
                release = json.loads(data.decode("utf-8"))
                for media in release["media"]:
                    return media["tracks"]
        except:
            Logger.error(
                "MusicBrainzWebHelper::get_tracks_payload(): %s", data)
        return []

    def lollypop_album_payload(self, payload):
        """
            Convert payload to Lollypop one
            @param payload as {}
            return {}
        """
        lollypop_payload = {}
        lollypop_payload["mbid"] = None
        lollypop_payload["name"] = payload["title"]
        lollypop_payload["uri"] = "mb:%s" % payload["id"]
        artists = []
        for artist in payload["artist-credit"]:
            artists.append(artist["name"])
        lollypop_payload["artists"] = ";".join(artists)
        lollypop_payload["track-count"] = payload["media"][0]["track-count"]
        lollypop_payload["artwork-uri"] = payload["id"]
        lollypop_payload["date"] = None
        return lollypop_payload

    def lollypop_track_payload(self, payload):
        """
            Convert payload to Lollypop one
            @param payload as {}
            @return {}
        """
        lollypop_payload = {}
        lollypop_payload["mbid"] = None
        lollypop_payload["name"] = payload["title"]
        lollypop_payload["uri"] = "mb:%s" % payload["id"]
        artists = []
        for artist in payload["artist-credit"]:
            artists.append(artist["name"])
        lollypop_payload["artists"] = ";".join(artists)
        lollypop_payload["discnumber"] = "1"
        lollypop_payload["tracknumber"] = payload["position"]
        lollypop_payload["duration"] = payload["length"]
        return lollypop_payload

#######################
# PRIVATE             #
#######################
